# RCAIDE/Library/Methods/Powertrain/Systems/append_payload_conditions.py
# 
# Created:  Jun 2024, M. Clarke  

from RCAIDE.Framework.Mission.Common     import   Conditions

# ---------------------------------------------------------------------------------------------------------------------- 
#  append_payload_conditions
# ----------------------------------------------------------------------------------------------------------------------    
def append_payload_conditions(payload, segment, bus):  
    """
    Initializes and appends empty payload conditions data structures to the segment state conditions.
    
    Parameters
    ----------
    payload : Payload
        The payload component for which conditions are being initialized.
    segment : Segment
        The mission segment in which the payload is operating.
    bus : ElectricalBus
        The electrical bus that powers the payload system.
    
    Returns
    -------
    None
        This function modifies the segment.state.conditions.energy dictionary in-place.
    
    Notes
    -----
    This function creates an empty Conditions object for the payload system within
    the segment's energy conditions dictionary, indexed by the bus tag and payload tag.
    
    The payload power consumption is initialized as a zero array with the same
    length as the segment's state vector. This will be updated during mission analysis
    based on the payload power requirements.
    
    See Also
    --------
    RCAIDE.Library.Methods.Powertrain.Systems.append_avionics_conditions
    """
    ones_row    = segment.state.ones_row 
    segment.state.conditions.energy[bus.tag][payload.tag]       = Conditions()
    segment.state.conditions.energy[bus.tag][payload.tag].power = 0 * ones_row(1)  
    return 



# RCAIDE/Methods/Energy/Propulsors/Electric_Rotor_Propulsor/compute_electric_rotor_performance.py
# 
# 
# Created:  Jul 2023, M. Clarke

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------
# RCAIDE imports   
from RCAIDE.Library.Methods.Powertrain.Modulators.Electronic_Speed_Controller.compute_esc_performance  import * 
from RCAIDE.Library.Methods.Powertrain.Converters.Motor.compute_motor_performance                      import *
from RCAIDE.Library.Methods.Powertrain.Converters.Rotor.compute_rotor_performance                      import * 

# pacakge imports  
import numpy as np 
from copy import deepcopy

# ----------------------------------------------------------------------------------------------------------------------
# compute_electric_rotor_performance
# ----------------------------------------------------------------------------------------------------------------------  
def compute_electric_rotor_performance(propulsor, state, center_of_gravity=[[0.0, 0.0, 0.0]]):
    """
    Computes the performance of an electric rotor propulsion system.
    
    Parameters
    ----------
    propulsor : RCAIDE.Library.Components.Propulsors.Electric_Rotor
        Electric rotor propulsor component with the following attributes:
            - tag : str
                Identifier for the propulsor
            - motor : Data
                Electric motor component
                - tag : str
                    Identifier for the motor
            - rotor : Data
                Rotor component
                    - tag : str
                        Identifier for the rotor
                    - origin : list of lists
                        Origin coordinates [[x, y, z]] [m]
            - electronic_speed_controller : Data
                ESC component
                    - tag : str
                        Identifier for the ESC
    state : RCAIDE.Framework.Mission.Common.State
        State object containing:
            - conditions : Data
                Flight conditions
                    - energy : Data
                        Energy conditions
                            - propulsors : dict
                                Propulsor energy conditions indexed by tag
                            - converters : dict
                                Converter energy conditions indexed by tag
                            - modulators : dict
                                Modulator energy conditions indexed by tag
            - ones_row : function
                Function to create array of ones with specified length
    center_of_gravity : list of lists, optional
        Center of gravity coordinates [[x, y, z]] [m]
        Default: [[0.0, 0.0, 0.0]]
    
    Returns
    -------
    thrust : numpy.ndarray
        Thrust force vector [N]
    moment : numpy.ndarray
        Moment vector [N·m]
    power : numpy.ndarray
        Mechanical power [W]
    power_elec : numpy.ndarray
        Electrical power [W]
    stored_results_flag : bool
        Flag indicating if results are stored
    stored_propulsor_tag : str
        Tag of the propulsor with stored results
    
    Notes
    -----
    This function computes the performance of an electric rotor propulsion system
    by linking the electronic speed controller (ESC), motor, and rotor. It performs 
    the following steps:
        * Sets the ESC throttle based on the propulsor throttle
        * Computes the ESC output voltage
        * Transfers the ESC output voltage to the motor
        * Computes the motor performance (angular velocity and torque)
        * Transfers the motor angular velocity to the rotor
        * Computes the rotor performance
        * Calculates the moment generated by the rotor thrust
        * Computes the ESC input current
        * Stores the results in the conditions data structure
    
    **Major Assumptions**
        * The ESC throttle directly controls the output voltage
        * The motor angular velocity directly drives the rotor
        * Moment is calculated relative to the aircraft center of gravity
        * Moment from motor rotation is not included
    
    See Also
    --------
    RCAIDE.Library.Methods.Powertrain.Converters.Motor.compute_motor_performance
    RCAIDE.Library.Methods.Powertrain.Converters.Rotor.compute_rotor_performance
    RCAIDE.Library.Methods.Powertrain.Modulators.Electronic_Speed_Controller
    """
     
    conditions                 = state.conditions    
    motor                      = propulsor.motor 
    rotor                      = propulsor.rotor 
    esc                        = propulsor.electronic_speed_controller   
    electric_rotor_conditions  = conditions.energy.propulsors[propulsor.tag]
    eta                        = electric_rotor_conditions.throttle
     
    conditions.energy.modulators[esc.tag].throttle         = eta 
    compute_voltage_out_from_throttle(esc,conditions)

    # Assign conditions to the rotor
    conditions.energy.converters[motor.tag].inputs.voltage = conditions.energy.modulators[esc.tag].outputs.voltage  
    compute_motor_performance(motor,conditions) 
    
    # Spin the rotor 
    conditions.energy.converters[rotor.tag].omega           = conditions.energy.converters[motor.tag].outputs.omega
    conditions.energy.converters[rotor.tag].throttle        = conditions.energy.modulators[esc.tag].throttle      
    conditions.energy.converters[rotor.tag].commanded_thrust_vector_angle =  conditions.energy.propulsors[propulsor.tag].commanded_thrust_vector_angle
    compute_rotor_performance(rotor,conditions)
 
    # Compute moment 
    moment_vector           = 0*state.ones_row(3)
    moment_vector[:,0]      = rotor.origin[0][0]  -  center_of_gravity[0][0] 
    moment_vector[:,1]      = rotor.origin[0][1]  -  center_of_gravity[0][1] 
    moment_vector[:,2]      = rotor.origin[0][2]  -  center_of_gravity[0][2]
    moment                  =  np.cross(moment_vector, conditions.energy.converters[rotor.tag].thrust)     
    
    # Detemine esc current 
    conditions.energy.modulators[esc.tag].outputs.current = conditions.energy.converters[motor.tag].inputs.current
    compute_current_in_from_throttle(esc,conditions)
    
    stored_results_flag     = True
    stored_propulsor_tag    = propulsor.tag 
    
    # compute total forces and moments from propulsor (future work would be to add moments from motors)
    electric_rotor_conditions.thrust      = conditions.energy.converters[rotor.tag].thrust 
    electric_rotor_conditions.power       = conditions.energy.converters[rotor.tag].power 
    electric_rotor_conditions.moment      = moment
    
    return electric_rotor_conditions.thrust,electric_rotor_conditions.moment,electric_rotor_conditions.power,conditions.energy.modulators[esc.tag].inputs.power,stored_results_flag,stored_propulsor_tag 
                
def reuse_stored_electric_rotor_data(propulsor,state,network,stored_propulsor_tag,center_of_gravity= [[0.0, 0.0,0.0]]):
    '''Reuses results from one propulsor for identical propulsors
    
    Assumptions: 
    N/A

    Source:
    N/A

    Inputs:  
    conditions           - operating conditions data structure    [-] 
    voltage              - system voltage                         [V]
    bus                  - bus                                    [-] 
    propulsors           - propulsor data structure               [-] 
    total_thrust         - thrust of propulsor group              [N]
    total_power          - power of propulsor group               [W]
    total_current        - current of propulsor group             [A]

    Outputs:  
    total_thrust         - thrust of propulsor group              [N]
    total_power          - power of propulsor group               [W]
    total_current        - current of propulsor group             [A] 
    
    Properties Used: 
    N.A.        
    ''' 
    conditions                 = state.conditions 
    motor                      = propulsor.motor 
    rotor                      = propulsor.rotor 
    esc                        = propulsor.electronic_speed_controller  
    motor_0                    = network.propulsors[stored_propulsor_tag].motor 
    rotor_0                    = network.propulsors[stored_propulsor_tag].rotor 
    esc_0                      = network.propulsors[stored_propulsor_tag].electronic_speed_controller
    
    conditions.energy.converters[motor.tag]        = deepcopy(conditions.energy.converters[motor_0.tag])
    conditions.energy.converters[rotor.tag]        = deepcopy(conditions.energy.converters[rotor_0.tag])
    conditions.energy.modulators[esc.tag]          = deepcopy(conditions.energy.modulators[esc_0.tag])
  
    thrust_vector           = conditions.energy.converters[rotor.tag].thrust 
    P_mech                  = conditions.energy.converters[rotor.tag].power 
    P_elec                  = conditions.energy.modulators[esc.tag].inputs.power    
    
    moment_vector           = 0*state.ones_row(3) 
    moment_vector[:,0]      = rotor.origin[0][0]  -  center_of_gravity[0][0] 
    moment_vector[:,1]      = rotor.origin[0][1]  -  center_of_gravity[0][1] 
    moment_vector[:,2]      = rotor.origin[0][2]  -  center_of_gravity[0][2]
    moment                  =  np.cross(moment_vector, thrust_vector)
     
    conditions.energy.propulsors[propulsor.tag].power             = P_mech  
    conditions.energy.propulsors[propulsor.tag].thrust            = thrust_vector  
    conditions.energy.propulsors[propulsor.tag].moment            = moment          
    return thrust_vector,moment,P_mech,P_elec


# RCAIDE/Library/Methods/Weights/Correlation_Buildups/Raymer/compute_vertical_tail_weight.py
# 
# 
# Created:  Sep 2024, M. Clarke

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------

# RCAIDE 
from RCAIDE.Framework.Core    import Units

# python imports 
import  numpy as  np
 
# ----------------------------------------------------------------------------------------------------------------------
#  Vertical Tail Weight 
# ----------------------------------------------------------------------------------------------------------------------
def compute_vertical_tail_weight(vehicle, wing, settings):
    """
    Calculates vertical tail weight using Raymer's empirical method.

    Parameters
    ----------
    vehicle : RCAIDE.Vehicle()
        Vehicle data structure containing:
            - mass_properties.max_takeoff : float
                Maximum takeoff weight [kg]
            - flight_envelope.ultimate_load : float
                Ultimate load factor
            - wings['main_wing'] : Data()
                Main wing properties:
                    - origin : array
                        Root location [m]
                    - aerodynamic_center : array
                        Location of aerodynamic center [m]
    wing : RCAIDE.Component()
        Vertical tail component containing:
            - t_tail : bool
                T-tail configuration flag
            - areas.reference : float
                Reference area [m^2]
            - origin : array
                Root location [m]
            - aerodynamic_center : array
                Location of aerodynamic center [m]
            - sweeps.quarter_chord : float
                Quarter chord sweep angle [rad]
            - aspect_ratio : float
                Aspect ratio
            - thickness_to_chord : float
                Thickness-to-chord ratio

    Returns
    -------
    tail_weight : float
        Weight of the vertical tail [kg]

    Notes
    -----
    This method implements Raymer's correlation for transport aircraft vertical
    tail weight estimation, accounting for geometry, loads, and configuration effects.

    **Major Assumptions**
        * T-tail configuration handled through multiplier
        * Correlation based on transport category aircraft data
        * If advanced composites are used, the vertical tail weight is reduced by 15%
    **Theory**
    The vertical tail weight is calculated using:
    .. math::
        W_{vt} = 0.0026(1 + H_t)^{0.225}W_{dg}^{0.556}N_{ult}^{0.536}L_t^{-0.5}S_{vt}^{0.5}K_z^{0.875}\cos(\Lambda)^{-1}AR^{0.35}(t/c)^{-0.5}

    where:
        * :math:`H_t` is T-tail factor (0 or 1)
        * :math:`W_{dg}` is design gross weight
        * :math:`N_{ult}` is ultimate load factor
        * :math:`L_t` is tail arm length
        * :math:`S_{vt}` is vertical tail area
        * :math:`K_z` is vertical tail arm
        * :math:`\Lambda` is quarter-chord sweep
        * :math:`AR` is aspect ratio
        * :math:`t/c` is thickness ratio

    References
    ----------
    [1] Raymer, D., "Aircraft Design: A Conceptual Approach", AIAA 
        Education Series, 2018. 

    See Also
    --------
    RCAIDE.Library.Methods.Mass_Properties.Weight_Buildups.Conventional.Transport.Raymer.compute_horizontal_tail_weight
    RCAIDE.Library.Methods.Mass_Properties.Weight_Buildups.Conventional.Transport.Raymer.compute_main_wing_weight
    RCAIDE.Library.Methods.Mass_Properties.Weight_Buildups.Conventional.Transport.Raymer.compute_operating_empty_weight
    """
    DG          = vehicle.mass_properties.max_takeoff / Units.lbs
    t_tail_flag = wing.t_tail
    wing_origin = wing.origin[0][0] / Units.ft
    wing_ac     = wing.aerodynamic_center[0] / Units.ft
    main_origin = vehicle.wings['main_wing'].origin[0][0] / Units.ft
    main_ac     = vehicle.wings['main_wing'].aerodynamic_center[0] / Units.ft
    Svt         = wing.areas.reference / Units.ft ** 2
    sweep       = wing.sweeps.quarter_chord
    Av          = wing.aspect_ratio
    t_c         = wing.thickness_to_chord 
    Nult        = vehicle.flight_envelope.ultimate_load
    
    H = 0
    if t_tail_flag:
        H = 1
    Lt = (wing_origin + wing_ac - main_origin - main_ac)
    Kz = Lt
    tail_weight = 0.0026 * (1 + H) ** 0.225 * DG ** 0.556 * Nult ** 0.536 \
                  * Lt ** (-0.5) * Svt ** 0.5 * Kz ** 0.875 * np.cos(sweep) ** (-1) * Av ** 0.35 * t_c ** (-0.5)
    if settings.advanced_composites:
        tail_weight *= 0.85
    
    return tail_weight * Units.lbs



# Magnesium.py
#
# Created: Jul, 2022, J. Smart
# Modified:

# -------------------------------------------------------------------------------
# Imports
# -------------------------------------------------------------------------------

from .Solid import Solid
from RCAIDE.Framework.Core import Units


# -------------------------------------------------------------------------------
# Aluminum 6061-T6 Solid Class
# ------------------------------------------------------------------------------- 
class Magnesium(Solid):
    """
    A class representing RZ5 magnesium alloy material properties per BS 2L.128 standard.

    Attributes
    ----------
    ultimate_tensile_strength : float
        Maximum tensile stress before failure in Pa (200e6)
    ultimate_shear_strength : float
        Maximum shear stress before failure in Pa (138e6)
    ultimate_bearing_strength : float
        Maximum bearing stress before failure in Pa (330e6)
    yield_tensile_strength : float
        Stress at which material begins to deform plastically in Pa (135e6)
    yield_shear_strength : float
        Shear stress at which material begins to deform plastically in Pa (138e6)
    yield_bearing_strength : float
        Bearing stress at which material begins to deform plastically in Pa (130e6)
    minimum_gage_thickness : float
        Minimum manufacturable thickness in m (0.0)
    density : float
        Material density in kg/m³ (1840)

    Notes
    -----
    This class implements material properties for RZ5 magnesium alloy, which is commonly 
    used in aerospace applications due to its high strength-to-weight ratio. The zero 
    value for minimum gage thickness indicates that this should be specified based on 
    specific manufacturing capabilities.

    **Definitions**
    
    'RZ5'
        A magnesium alloy designation per British Standard BS 2L.128, containing zinc 
        and rare earth elements for improved strength
    
    'Ultimate Strength'
        The maximum stress that a material can withstand before failure
    
    'Yield Strength'
        The stress at which a material begins to deform plastically

    References
    ----------
    [1] MatWeb. (n.d.). Magnesium Elektron Elektron® RZ5 Magnesium Casting Alloy, UNS M16410. Magnesium elektron elektron® rz5 magnesium casting alloy, uns m16410. https://www.matweb.com/search/datasheet.aspx?matguid=f473a6bbcabe4fd49199c2cef7205664 
    """

    def __defaults__(self):
        """Sets material properties at instantiation.

        Assumptions:
        None

        Source:
        N/A

        Inputs:
        N/A

        Outputs:
        N/A

        Properties Used:
        None
        """

        self.ultimate_tensile_strength  = 200e6 * Units.Pa
        self.ultimate_shear_strength    = 138e6 * Units.Pa
        self.ultimate_bearing_strength  = 330e6 * Units.Pa
        self.yield_tensile_strength     = 135e6 * Units.Pa
        self.yield_shear_strength       = 138e6 * Units.Pa
        self.yield_bearing_strength     = 130e6 * Units.Pa
        self.minimum_gage_thickness     = 0.0   * Units.m
        self.density                    = 1840. * Units['kg/(m**3)']



# Paint.py
#
# Created: Jul, 2017, J. Smart
# Modified: Apr, 2018, J. Smart

#-------------------------------------------------------------------------------
# Imports
#-------------------------------------------------------------------------------

from .Solid import Solid
from RCAIDE.Framework.Core import Units

#-------------------------------------------------------------------------------
# Paint and/or Vinyl Surface Convering Solid Class
#-------------------------------------------------------------------------------

class Paint(Solid):
    """
    A class representing paint and vinyl surface coating material properties.

    Attributes
    ----------
    ultimate_tensile_strength : float
        Maximum tensile stress before failure in Pa (0.0)
    ultimate_shear_strength : float
        Maximum shear stress before failure in Pa (0.0)
    ultimate_bearing_strength : float
        Maximum bearing stress before failure in Pa (0.0)
    yield_tensile_strength : float
        Stress at which material begins to deform plastically in Pa (0.0)
    yield_shear_strength : float
        Shear stress at which material begins to deform plastically in Pa (0.0)
    yield_bearing_strength : float
        Bearing stress at which material begins to deform plastically in Pa (0.0)
    minimum_gage_thickness : float
        Minimum applicable thickness in m (150e-6)
    density : float
        Material density in kg/m³ (1800)

    Notes
    -----
    This class implements material properties for paint and vinyl surface coatings. 
    The zero values for strength properties reflect that these coatings are non 
    load-bearing elements. The minimum gage thickness represents a typical 
    coating thickness for aerospace applications.

    **Definitions**
    
    'Surface Coating'
        A material applied to the surface of another material to provide protection, 
        decoration, or other functional properties
    
    'Minimum Gage Thickness'
        The minimum thickness required for adequate surface coverage and protection
    """

    def __defaults__(self):
        """Sets material properties at instantiation.

        Assumptions:
        None

        Source:
        N/A

        Inputs:
        N/A

        Outputs:
        N/A

        Properties Used:
        None
        """

        self.ultimate_tensile_strength  = 0.0       * Units.Pa
        self.ultimate_shear_strength    = 0.0       * Units.Pa
        self.ultimate_bearing_strength  = 0.0       * Units.Pa
        self.yield_tensile_strength     = 0.0       * Units.Pa
        self.yield_shear_strength       = 0.0       * Units.Pa
        self.yield_bearing_strength     = 0.0       * Units.Pa
        self.minimum_gage_thickness     = 150e-6    * Units.m
        self.density                    = 1800.     * Units['kg/(m**3)']



# RCAIDE/Library/Components/Propulsors/Converters/PMSM_Motor.py
# 
# 
# Created:  Jan 2025, M. Clarke, M. Guidotti

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------   
# RCAIDE imports  
from RCAIDE.Framework.Core import Data
from .Converter  import Converter
from RCAIDE.Library.Methods.Powertrain.Converters.Motor.append_motor_conditions import  append_motor_conditions

# ----------------------------------------------------------------------------------------------------------------------
#  PMSM_Motor  
# ----------------------------------------------------------------------------------------------------------------------           
class PMSM_Motor(Converter):
    """
    Permanent Magnet Synchronous Motor (PMSM) Component Class

    This class models a PMSM motor for electric propulsion systems. It inherits from the base Converter class
    and implements motor-specific attributes and methods.

    Attributes
    ----------
    tag : str
        Identifier for the motor component, defaults to 'PMSM_motor'

    speed_constant : float
        Motor speed constant [rpm/V]
    stator_inner_diameter : float
        Inner diameter of the stator [m]
    stator_outer_diameter : float
        Outer diameter of the stator [m]
    winding_factor : float
        Winding factor of the motor [-]
    resistance : float
        Motor winding resistance [Ω]
    motor_stack_length : float
        Length of the motor stack [m]
    number_of_turns : int
        Number of winding turns [-]
    length_of_path : float
        Length of the magnetic path [m]
    mu_0 : float
        Permeability of free space [N/A^2]
    mu_r : float
        Relative permeability of magnetic material [N/A^2]
    thermal_conductivity : float
        Thermal conductivity of magnetic material [W/m*K]
    Delta_T : float
        Temperature difference between inner and outer stator surfaces [K]
    characteristic_length_of_flow : float
        Characteristic length for thermal calculations [m]
    thermal_conductivity_fluid : float
        Thermal conductivity of cooling fluid [W/m*K]
    length_of_conductive_path : float
        Length of thermal conductive path [m]
    Re_cooling_flow : float
        Reynolds number of cooling flow [-]
    Re_airgap : float
        Reynolds number of airgap flow [-]
    Prandtl_number : float
        Prandtl number of fluid [-]
    height_of_duct : float
        Height of cooling duct [m]
    width_of_duct : float
        Width of cooling duct [m]
    hydraulic_diameter_of_duct : float
        Hydraulic diameter of cooling duct [m]
    length_of_channel : float
        Length of cooling channel [m]
    volume_flow_rate_of_fluid : float
        Volume flow rate of cooling fluid [m^3/s]
    density_of_fluid : float
        Density of cooling fluid [kg/m^3]
    velocity_of_fluid : float
        Velocity of cooling fluid [m/s]
    Taylor_number : float
        Taylor number for flow calculations [-]
    axial_gap_to_radius_of_rotor : float
        Ratio of axial gap to rotor radius [-]
    Conduction_laminar_flow : bool
        Flag for laminar conduction flow
    Convection_laminar_flow : bool
        Flag for laminar convection flow

    Notes
    -----
    The PMSM motor model includes detailed thermal and electromagnetic parameters for accurate 
    performance modeling. Default values are provided for all parameters but should be updated
    based on specific motor designs.

    **Major Assumptions**
        * Linear magnetic properties
        * Uniform temperature distribution in components
        * Simplified thermal model
        * No saturation effects considered

    See Also
    --------
    RCAIDE.Library.Components.Powertrain.Converters
    """      
    def __defaults__(self):
        """
        """           
        self.tag                           = 'PMSM_motor' 
        # Input data from Datasheet      
        self.speed_constant                = 6.56                        # [rpm/V]        speed constant
        self.stator_inner_diameter         = 0.16                        # [m]            stator inner diameter
        self.stator_outer_diameter         = 0.348                       # [m]            stator outer diameter
        self.gearbox                       = Data()
        self.gearbox.gear_ratio            = 1.0 
      
        # Input data from Literature      
        self.winding_factor                = 0.95                        # [-]            winding factor

        # Input data from Assumptions
        self.resistance                    = 0.002                       # [Ω]            resistance
        self.motor_stack_length            = 0.1140                      # [m]            (It should be around 0.14 m) motor stack length 
        self.number_of_turns               = 80                          # [-]            number of turns  
        self.length_of_path                = 0.4                         # [m]            length of the path  
        self.mu_0                          = 1.256637061e-6              # [N/A**2]       permeability of free space
        self.mu_r                          = 1005                        # [N/A**2]       relative permeability of the magnetic material 
        self.thermal_conductivity          = 200                         # [W/m*K]        thermal conductivity of the magnetic material
        self.Delta_T                       = 10                          # [K]            temperature difference between the inner and outer surfaces of the stator
        self.characteristic_length_of_flow = 0.01                    # [m]            characteristic length of the flow
        self.thermal_conductivity_fluid    = 0.026                      # [W/m*K]        thermal conductivity of the fluid
        self.length_of_conductive_path     = 0.4                         # [m]            length of the conductive path  
        self.Re_cooling_flow               = 100000                      # [-]            Reynolds number of the coolingflow
        self.Re_airgap                     = 100000                      # [-]            Reynolds number of the flow in the airgap
        self.Prandtl_number                = 0.708                       # [-]            Prandtl number of the flow
        self.height_of_duct                = 0.005                       # [m]            height of the duct
        self.width_of_duct                 = 0.005                       # [m]            width of the duct
        self.hydraulic_diameter_of_duct    = 0.005                      # [m]            hydraulic diameter of the duct
        self.length_of_channel             = 0.005                       # [m]            length of the channel
        self.volume_flow_rate_of_fluid     = 0.005                       # [m**3/s]       volume flow rate of the fluid
        self.density_of_fluid              = 1000                        # [kg/m**3]      density of the fluid
        self.velocity_of_fluid             = 0.005                       # [m/s]          velocity of the fluid
        self.Taylor_number                 = 20                          # [-]            Taylor number 
        self.axial_gap_to_radius_of_rotor  = 0.01                     # [-]            ratio of the axial gap to the radius of the rotor 
        self.inverse_calculation           = False
        self.Conduction_laminar_flow       = True                        # [-]            True if the flow is laminar, False if the flow is turbulent
        self.Convection_laminar_flow       = True                        # [-]            True if the flow is laminar, False if the flow is turbulent
        
    def append_operating_conditions(self,segment,energy_conditions,noise_conditions=None): 
        append_motor_conditions(self,segment,energy_conditions)
        return
    


# RCAIDE/Library/Methods/Mass_Properties/Weight_Buildups/Conventional/Transport/Raymer/__init__.py
# 

"""
Collection of aircraft weight estimation methods based on Raymer's empirical correlations for a transport aircraft. 
This module provides component weight estimation functions including wing, tail, fuselage, landing gear, 
and systems weights. The correlations are primarily derived from historical aircraft data as presented in 
Raymer's "Aircraft Design: A Conceptual Approach".

See Also
--------
RCAIDE.Library.Methods.Mass_Properties.Weight_Buildups.Conventional.Transport.FLOPS
RCAIDE.Library.Methods.Mass_Properties.Weight_Buildups.Conventional.General_Aviation
"""

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------
from .compute_operating_empty_weight import compute_operating_empty_weight
from .compute_main_wing_weight import compute_main_wing_weight
from .compute_horizontal_tail_weight import compute_horizontal_tail_weight
from .compute_vertical_tail_weight import compute_vertical_tail_weight
from .compute_fuselage_weight import compute_fuselage_weight
from .compute_landing_gear_weight import compute_landing_gear_weight
from .compute_propulsion_system_weight import *
from .compute_systems_weight import *


# RCAIDE/Library/Methods/Powertrain/Systems/compute_payload_power_draw.py
# 
# Created:  Jul 2024, RCAIDE Team 

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------    
# package imports  
def compute_payload_power_draw(payload, bus, conditions):
    """
    Computes the power draw of a payload component on an electrical bus.
    
    Parameters
    ----------
    payload : RCAIDE.Library.Components.Systems.Payload
        Payload component with the following attributes:
            - tag : str
                Identifier for the payload
            - power_draw : float
                Power required by the payload [W]
    bus : RCAIDE.Library.Components.Systems.Electrical_Bus
        Electrical bus component with the following attributes:
            - tag : str
                Identifier for the electrical bus
            - power_split_ratio : float
                Ratio of power allocation to this bus
            - efficiency : float
                Efficiency of the bus [0-1]
    conditions : RCAIDE.Framework.Mission.Common.Conditions
        Flight conditions with:
            - energy : dict
                Energy conditions indexed by component tag
                    - [bus.tag] : Data
                        Bus-specific conditions
                            - [payload.tag] : Data
                                Payload-specific conditions
                                - power : numpy.ndarray
                                    Power draw of the payload [W]
                            - power_draw : numpy.ndarray
                                Total power draw on the bus [W]
    
    Returns
    -------
    None
    
    Notes
    -----
    This function calculates the power draw of a payload component on an electrical bus
    and updates the total power draw on the bus. The power draw is adjusted by the
    bus power split ratio and efficiency.
    
    **Major Assumptions**
        * Constant payload power draw
        * Constant bus efficiency
        * Power draw is independent of flight conditions
    
    See Also
    --------
    RCAIDE.Library.Methods.Powertrain.Systems.compute_avionics_power_draw
    """
    bus_conditions                = conditions.energy[bus.tag]
    payload_conditions            = bus_conditions[payload.tag]    
    payload_conditions.power[:,0] = payload.power_draw  
    bus_conditions.power_draw     += payload_conditions.power*bus.power_split_ratio /bus.efficiency
    return 


# RCAIDE/Library/Attributes/Planets/__init__.py
# 

"""
This module provides planetary environment definitions for RCAIDE simulations. It includes 
base classes and specific implementation for Earth.

The module supports atmospheric calculations, gravitational effects, and other planetary 
environmental factors needed for aerospace vehicle analysis and simulation.

See Also
--------
RCAIDE.Library.Attributes.Atmospheres : Detailed atmospheric models
"""

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------

from .Planet import Planet
from .Earth  import Earth


# RCAIDE/Library/Methods/Powertrain/Propulsors/Constant_Speed_Internal_Combustion_Engine/design_constant_speed_internal_combustion_engine.py
# 
# Created:  Mar 2025, M. Clarke

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------

# RCAIDE Imports
import RCAIDE 
from RCAIDE.Library.Methods.Powertrain.Converters.Rotor  import design_propeller  
from RCAIDE.Library.Methods.Powertrain                   import setup_operating_conditions 

# Python package imports
import numpy as np

# ----------------------------------------------------------------------------------------------------------------------
#  Design Electric Rotor 
# ----------------------------------------------------------------------------------------------------------------------
def design_constant_speed_internal_combustion_engine(ICE_CS, number_of_stations=20):
    """
    Sizes the propeller of a propeller-driven constant speed internal combustion engine and computes
    sea level static performance.
    
    Parameters
    ----------
    ICE_CS : RCAIDE.Library.Components.Propulsors.Constant_Speed_Internal_Combustion_Engine
        Constant speed internal combustion engine propulsor component with the following attributes:
            - tag : str
                Identifier for the propulsor
            - propeller : Data
                Propeller component to be designed
    number_of_stations : int, optional
        Number of radial stations for propeller blade discretization
        Default: 20
    
    Returns
    -------
    None
        Results are stored in the ICE_CS object:
            - sealevel_static_thrust : float
                Sea level static thrust [N]
            - sealevel_static_power : float
                Sea level static power [W]
    
    Notes
    -----
    This function performs two main tasks:
        1. Designs the propeller using the design_propeller function with the specified
           number of stations
        2. Computes the sea level static performance (thrust and power) of the engine-propeller
           combination at full throttle
    
    The sea level static performance is calculated by:
        - Setting up atmospheric conditions at sea level
        - Creating a low-speed operating state (1% of sea level speed of sound)
        - Setting the throttle to maximum (1.0)
        - Computing the performance at these conditions
    
    Unlike variable speed internal combustion engines, the constant speed variant maintains
    a fixed RPM and varies the propeller blade pitch to control thrust. This is common in
    generator applications and some aircraft engines where constant speed operation is
    beneficial for efficiency or noise reasons.
    
    **Major Assumptions**
        * US Standard Atmosphere 1976 is used for atmospheric properties
        * Sea level static conditions are approximated with a very low velocity (1% of speed of sound)
        * Full throttle (throttle = 1.0) is used for sea level static performance
    
    See Also
    --------
    RCAIDE.Library.Methods.Powertrain.Converters.Rotor.design_propeller
    RCAIDE.Library.Methods.Powertrain.setup_operating_conditions
    """
    
    # Step 1 Design the Propeller  
    design_propeller(ICE_CS.propeller,number_of_stations = 20) 
     
    # Static Sea Level Thrust   
    atmosphere            = RCAIDE.Framework.Analyses.Atmospheric.US_Standard_1976() 
    atmo_data_sea_level   = atmosphere.compute_values(0.0,0.0)   
    V                     = atmo_data_sea_level.speed_of_sound[0][0]*0.01 
    operating_state       = setup_operating_conditions(ICE_CS, altitude = 0,velocity_range=np.array([V]))  
    operating_state.conditions.energy.propulsors[ICE_CS.tag].throttle[:,0] = 1.0  
    sls_T,_,sls_P,_,_,_               = ICE_CS.compute_performance(operating_state) 
    ICE_CS.sealevel_static_thrust        = sls_T[0][0]
    ICE_CS.sealevel_static_power         = sls_P[0][0]
    return 


# RCAIDE/Methods/Energy/Propulsors/Networks/Turboprop/compute_turboprop_performance.py
# 
# 
# Created:  Jul 2023, M. Clarke

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------
# RCAIDE imports      

from RCAIDE.Framework.Core                                             import Data 
from RCAIDE.Library.Methods.Powertrain.Converters.Ram                  import compute_ram_performance
from RCAIDE.Library.Methods.Powertrain.Converters.Combustor            import compute_combustor_performance
from RCAIDE.Library.Methods.Powertrain.Converters.Compressor           import compute_compressor_performance
from RCAIDE.Library.Methods.Powertrain.Converters.Turbine              import compute_turbine_performance
from RCAIDE.Library.Methods.Powertrain.Converters.Expansion_Nozzle     import compute_expansion_nozzle_performance 
from RCAIDE.Library.Methods.Powertrain.Converters.Compression_Nozzle   import compute_compression_nozzle_performance
from RCAIDE.Library.Methods.Powertrain.Propulsors.Turboprop            import compute_thrust
 
# python imports 
from   copy import deepcopy
import numpy as np

# ----------------------------------------------------------------------------------------------------------------------
# compute_turboprop_performance
# ---------------------------------------------------------------------------------------------------------------------- 
def compute_turboprop_performance(turboprop, state, center_of_gravity=[[0.0, 0.0, 0.0]]):
    """
    Computes the performance of a turboprop engine by analyzing the thermodynamic cycle.
    
    Parameters
    ----------
    turboprop : RCAIDE.Library.Components.Propulsors.Turboprop
        Turboprop engine component with the following attributes:
            - tag : str
                Identifier for the turboprop
            - working_fluid : Data
                Working fluid properties object
            - ram : Data
                Ram component
                    - tag : str
                        Identifier for the ram
            - inlet_nozzle : Data
                Inlet nozzle component
                    - tag : str
                        Identifier for the inlet nozzle
            - compressor : Data
                Compressor component
                    - tag : str
                        Identifier for the compressor
                    - motor : Data, optional
                        Electric motor component
                    - generator : Data, optional
                        Electric generator component
                    - design_angular_velocity : float
                        Design angular velocity [rad/s]
            - combustor : Data
                Combustor component
                    - tag : str
                        Identifier for the combustor
                    - fuel_data : Data
                        Fuel properties
                        - specific_energy : float
                            Fuel specific energy [J/kg]
            - high_pressure_turbine : Data
                High pressure turbine component
                    - tag : str
                        Identifier for the high pressure turbine
            - low_pressure_turbine : Data
                Low pressure turbine component
                    - tag : str
                        Identifier for the low pressure turbine
            - core_nozzle : Data
                Core nozzle component
                    - tag : str
                        Identifier for the core nozzle
            - reference_temperature : float
                Reference temperature for mass flow scaling [K]
            - reference_pressure : float
                Reference pressure for mass flow scaling [Pa]
            - compressor_nondimensional_massflow : float
                Non-dimensional mass flow parameter [kg·√K/(s·Pa)]
            - origin : list of lists
                Origin coordinates [[x, y, z]] [m]
    state : RCAIDE.Framework.Mission.Common.State
        State object containing:
            - conditions : Data
                Flight conditions
                    - freestream : Data
                        Freestream properties
                            - velocity : numpy.ndarray
                                Freestream velocity [m/s]
                            - temperature : numpy.ndarray
                                Freestream temperature [K]
                            - pressure : numpy.ndarray
                                Freestream pressure [Pa]
                    - noise : Data
                        Noise conditions
                            - propulsors : dict
                                Propulsor noise conditions indexed by tag
                    - energy : Data
                        Energy conditions
                            - propulsors : dict
                                Propulsor energy conditions indexed by tag
                            - converters : dict
                                Converter energy conditions indexed by tag
                            - hybrid_power_split_ratio : float
                                Ratio of power split for hybrid systems
            - numerics : Data
                Numerical properties
                    - time : Data
                        Time properties
                            - differentiate : list
                                List of differentiation methods
    center_of_gravity : list of lists, optional
        Center of gravity coordinates [[x, y, z]] [m]
        Default: [[0.0, 0.0, 0.0]]
    
    Returns
    -------
    thrust_vector : numpy.ndarray
        Thrust force vector [N]
    moment : numpy.ndarray
        Moment vector [N·m]
    power : numpy.ndarray
        Shaft power output [W]
    power_elec : numpy.ndarray
        Electrical power input/output [W]
    stored_results_flag : bool
        Flag indicating if results are stored
    stored_propulsor_tag : str
        Tag of the turboprop with stored results
    
    Notes
    -----
    This function computes the performance of a turboprop engine by sequentially analyzing
    each component in the engine's thermodynamic cycle. It links the output conditions of
    each component to the input conditions of the next component in the flow path.
    
    The function follows this sequence:
        1. Set working fluid properties
        2. Compute ram performance
        3. Compute inlet nozzle performance
        4. Compute compressor performance
        5. Compute combustor performance
        6. Compute high pressure turbine performance
        7. Compute low pressure turbine performance
        8. Compute core nozzle performance
        9. Compute thrust and power output
        10. Calculate efficiencies
        11. Handle electrical power generation/consumption if applicable
    
    **Major Assumptions**
        * Steady state operation
        * One-dimensional flow through components
        * Adiabatic components except for the combustor
        * Perfect gas behavior with variable properties
    
    References
    ----------
    [1] Mattingly, J.D., "Elements of Gas Turbine Propulsion", 2nd Edition, AIAA Education Series, 2005. https://soaneemrana.org/onewebmedia/ELEMENTS%20OF%20GAS%20TURBINE%20PROPULTION2.pdf
    
    See Also
    --------
    RCAIDE.Library.Methods.Powertrain.Propulsors.Turboprop.compute_thrust
    """ 
    conditions               = state.conditions 
    noise_conditions         = conditions.noise.propulsors[turboprop.tag]  
    turboprop_conditions     = conditions.energy.propulsors[turboprop.tag]
    U0                       = conditions.freestream.velocity
    T                        = conditions.freestream.temperature
    P                        = conditions.freestream.pressure 
    ram                      = turboprop.ram
    inlet_nozzle             = turboprop.inlet_nozzle
    compressor               = turboprop.compressor
    combustor                = turboprop.combustor
    high_pressure_turbine    = turboprop.high_pressure_turbine
    low_pressure_turbine     = turboprop.low_pressure_turbine
    core_nozzle              = turboprop.core_nozzle 
    ram_conditions           = conditions.energy.converters[ram.tag]     
    inlet_nozzle_conditions  = conditions.energy.converters[inlet_nozzle.tag]
    core_nozzle_conditions   = conditions.energy.converters[core_nozzle.tag] 
    compressor_conditions    = conditions.energy.converters[compressor.tag]  
    combustor_conditions     = conditions.energy.converters[combustor.tag]
    lpt_conditions           = conditions.energy.converters[low_pressure_turbine.tag]
    hpt_conditions           = conditions.energy.converters[high_pressure_turbine.tag] 
     
    # Step 1: Set the working fluid to determine the fluid properties
    ram.working_fluid                                     = turboprop.working_fluid

    # Step 2: Compute flow through the ram , this computes the necessary flow quantities and stores it into conditions
    compute_ram_performance(ram,conditions)

    # Step 3: link inlet nozzle to ram 
    inlet_nozzle_conditions.inputs.stagnation_temperature = ram_conditions.outputs.stagnation_temperature
    inlet_nozzle_conditions.inputs.stagnation_pressure    = ram_conditions.outputs.stagnation_pressure
    inlet_nozzle_conditions.inputs.static_temperature     = ram_conditions.outputs.static_temperature
    inlet_nozzle_conditions.inputs.static_pressure        = ram_conditions.outputs.static_pressure
    inlet_nozzle_conditions.inputs.mach_number            = ram_conditions.outputs.mach_number
    inlet_nozzle.working_fluid                            = ram.working_fluid

    # Step 4: Compute flow through the inlet nozzle
    compute_compression_nozzle_performance(inlet_nozzle,conditions)      

    # Step 5: Link low pressure compressor to the inlet nozzle 
    compressor_conditions.inputs.stagnation_temperature   = inlet_nozzle_conditions.outputs.stagnation_temperature
    compressor_conditions.inputs.stagnation_pressure      = inlet_nozzle_conditions.outputs.stagnation_pressure
    compressor_conditions.inputs.static_temperature       = inlet_nozzle_conditions.outputs.static_temperature
    compressor_conditions.inputs.static_pressure          = inlet_nozzle_conditions.outputs.static_pressure
    compressor_conditions.inputs.mach_number              = inlet_nozzle_conditions.outputs.mach_number  
    compressor.working_fluid                              = inlet_nozzle.working_fluid
    compressor.nondimensional_massflow                    = turboprop.compressor_nondimensional_massflow
    compressor_conditions.reference_temperature           = turboprop.reference_temperature
    compressor_conditions.reference_pressure              = turboprop.reference_pressure
    
    # Step 6: Compute flow through the low pressure compressor
    compute_compressor_performance(compressor,conditions)
    
    # Step 7: Link the combustor to the high pressure compressor
    combustor_conditions.inputs.stagnation_temperature    = compressor_conditions.outputs.stagnation_temperature
    combustor_conditions.inputs.stagnation_pressure       = compressor_conditions.outputs.stagnation_pressure
    combustor_conditions.inputs.static_temperature        = compressor_conditions.outputs.static_temperature
    combustor_conditions.inputs.static_pressure           = compressor_conditions.outputs.static_pressure
    combustor_conditions.inputs.mach_number               = compressor_conditions.outputs.mach_number  
    combustor.working_fluid                               = compressor.working_fluid 
    
    # Step 8: Compute flow through the high pressor compressor 
    compute_combustor_performance(combustor,conditions)
    
    #link the high pressure turbione to the combustor 
    hpt_conditions.inputs.stagnation_temperature          = combustor_conditions.outputs.stagnation_temperature
    hpt_conditions.inputs.stagnation_pressure             = combustor_conditions.outputs.stagnation_pressure
    hpt_conditions.inputs.fuel_to_air_ratio               = combustor_conditions.outputs.fuel_to_air_ratio 
    hpt_conditions.inputs.static_temperature              = combustor_conditions.outputs.static_temperature
    hpt_conditions.inputs.static_pressure                 = combustor_conditions.outputs.static_pressure
    hpt_conditions.inputs.mach_number                     = combustor_conditions.outputs.mach_number 
    hpt_conditions.inputs.compressor                      = compressor_conditions.outputs 
    high_pressure_turbine.working_fluid                   = combustor.working_fluid
    hpt_conditions.inputs.bypass_ratio                    = 0.0
    
    compute_turbine_performance(high_pressure_turbine,conditions)
    
    #link the low pressure turbine to the high pressure turbine 
    lpt_conditions.inputs.stagnation_temperature          = hpt_conditions.outputs.stagnation_temperature
    lpt_conditions.inputs.stagnation_pressure             = hpt_conditions.outputs.stagnation_pressure 
    lpt_conditions.inputs.static_temperature              = hpt_conditions.outputs.static_temperature
    lpt_conditions.inputs.static_pressure                 = hpt_conditions.outputs.static_pressure 
    lpt_conditions.inputs.mach_number                     = hpt_conditions.outputs.mach_number     
    lpt_conditions.inputs.compressor                      = Data()
    lpt_conditions.inputs.compressor.work_done            = 0.0   
    lpt_conditions.inputs.compressor.external_shaft_work_done = 0.0
    lpt_conditions.inputs.bypass_ratio                    = 0.0 
    lpt_conditions.inputs.fuel_to_air_ratio               = combustor_conditions.outputs.fuel_to_air_ratio 
    low_pressure_turbine.working_fluid                    = high_pressure_turbine.working_fluid    
     
    compute_turbine_performance(low_pressure_turbine,conditions)
    
    #link the core nozzle to the low pressure turbine
    core_nozzle_conditions.inputs.stagnation_temperature  = lpt_conditions.outputs.stagnation_temperature
    core_nozzle_conditions.inputs.stagnation_pressure     = lpt_conditions.outputs.stagnation_pressure
    core_nozzle_conditions.inputs.static_temperature      = lpt_conditions.outputs.static_temperature
    core_nozzle_conditions.inputs.static_pressure         = lpt_conditions.outputs.static_pressure  
    core_nozzle_conditions.inputs.mach_number             = lpt_conditions.outputs.mach_number   
    core_nozzle.working_fluid                             = low_pressure_turbine.working_fluid 
    
    #flow through the core nozzle
    compute_expansion_nozzle_performance(core_nozzle,conditions)

    # compute the thrust using the thrust component
    
    turboprop_conditions.total_temperature_reference      = compressor_conditions.inputs.stagnation_temperature
    turboprop_conditions.total_pressure_reference         = compressor_conditions.inputs.stagnation_pressure 

    # Compute the power
    compute_thrust(turboprop,conditions) 

    # Compute forces and moments
    moment_vector      = 0*state.ones_row(3)
    thrust_vector      = 0*state.ones_row(3)
    thrust_vector[:,0] = turboprop_conditions.thrust[:,0]
    moment_vector[:,0] = turboprop.origin[0][0] -   center_of_gravity[0][0] 
    moment_vector[:,1] = turboprop.origin[0][1]  -  center_of_gravity[0][1] 
    moment_vector[:,2] = turboprop.origin[0][2]  -  center_of_gravity[0][2]
    M                  = np.cross(moment_vector, thrust_vector)   
    moment             = M 
    power              = turboprop_conditions.power 
  
    # compute efficiencies 
    mdot_air_core                                  = turboprop_conditions.core_mass_flow_rate 
    fuel_enthalpy                                  = combustor.fuel_data.specific_energy 
    mdot_fuel                                      = turboprop_conditions.fuel_flow_rate   
    h_e_c                                          = core_nozzle_conditions.outputs.static_enthalpy
    h_0                                            = turboprop.working_fluid.compute_cp(T,P) * T 
    h_t4                                           = combustor_conditions.outputs.stagnation_enthalpy
    h_t3                                           = compressor_conditions.outputs.stagnation_enthalpy 
    turboprop_conditions.overall_efficiency        = thrust_vector* U0 / (mdot_fuel * fuel_enthalpy)  
    turboprop_conditions.thermal_efficiency        = 1 - ((mdot_air_core +  mdot_fuel)*(h_e_c -  h_0) + mdot_fuel *h_0)/((mdot_air_core +  mdot_fuel)*h_t4 - mdot_air_core *h_t3)   
    compressor_conditions.omega                    = compressor.design_angular_velocity * turboprop_conditions.throttle 
    
    # compute electrical power if generated/supplied  
    power_elec = 0*state.ones_row(1)
    if compressor.motor != None and  len(state.numerics.time.differentiate) > 0: 
        compressor_motor_conditions                 = conditions.energy.converters[compressor.motor.tag] 
        compressor_motor_conditions.outputs.power   = power *conditions.energy.hybrid_power_split_ratio   
        compressor_motor_conditions.outputs.omega   = compressor_conditions.omega
        compressor_motor_conditions.outputs.torque  = compressor_motor_conditions.outputs.power / compressor_motor_conditions.outputs.omega   
        power_elec =  compressor_motor_conditions.outputs.power  
    
    if compressor.generator != None and len(state.numerics.time.differentiate) > 0: 
        compressor_generator_conditions                = conditions.energy.converters[compressor.generator.tag] 
        compressor_generator_conditions.inputs.power   = power *conditions.energy.hybrid_power_split_ratio  
        compressor_generator_conditions.inputs.omega   = compressor_conditions.omega
        compressor_generator_conditions.outputs.torque = compressor_generator_conditions.outputs.power / compressor_generator_conditions.outputs.omega  
        power_elec =  compressor_generator_conditions.inputs.power  
            
    # Store data
    core_nozzle_res = Data(
                exit_static_temperature             = core_nozzle_conditions.outputs.static_temperature,
                exit_static_pressure                = core_nozzle_conditions.outputs.static_pressure,
                exit_stagnation_temperature         = core_nozzle_conditions.outputs.stagnation_temperature,
                exit_stagnation_pressure            = core_nozzle_conditions.outputs.static_pressure,
                exit_velocity                       = core_nozzle_conditions.outputs.velocity
            )
  
    noise_conditions.core_nozzle   = core_nozzle_res  
    
    # Pack results    
    stored_results_flag    = True
    stored_propulsor_tag   = turboprop.tag
    return thrust_vector,moment,power,power_elec,stored_results_flag,stored_propulsor_tag 

def reuse_stored_turboprop_data(turboprop,state,network,stored_propulsor_tag,center_of_gravity= [[0.0, 0.0,0.0]]):
    '''Reuses results from one turboprop for identical propulsors
    
    Assumptions: 
    N/A

    Source:
    N/A

    Inputs:  
    conditions           - operating conditions data structure     [-]  
    fuel_line            - fuelline                                [-] 
    turboprop            - turboprop data structure              [-] 
    total_power          - power of turboprop group               [W] 

    Outputs:  
    total_power          - power of turboprop group               [W] 
    
    Properties Used: 
    N.A.        
    ''' 
    # unpack
    conditions                  = state.conditions 
    ram                         = turboprop.ram
    inlet_nozzle                = turboprop.inlet_nozzle 
    compressor                  = turboprop.compressor 
    combustor                   = turboprop.combustor
    high_pressure_turbine       = turboprop.high_pressure_turbine
    low_pressure_turbine        = turboprop.low_pressure_turbine
    core_nozzle                 = turboprop.core_nozzle
    ram_0                       = network.propulsors[stored_propulsor_tag].ram
    inlet_nozzle_0              = network.propulsors[stored_propulsor_tag].inlet_nozzle 
    compressor_0                = network.propulsors[stored_propulsor_tag].compressor 
    combustor_0                 = network.propulsors[stored_propulsor_tag].combustor
    high_pressure_turbine_0     = network.propulsors[stored_propulsor_tag].high_pressure_turbine
    low_pressure_turbine_0      = network.propulsors[stored_propulsor_tag].low_pressure_turbine
    core_nozzle_0               = network.propulsors[stored_propulsor_tag].core_nozzle

    # deep copy results 
    conditions.energy.propulsors[turboprop.tag]                = deepcopy(conditions.energy.propulsors[stored_propulsor_tag])
    conditions.noise.propulsors[turboprop.tag]                 = deepcopy(conditions.noise.propulsors[stored_propulsor_tag]) 
    conditions.energy.converters[ram.tag]                      = deepcopy(conditions.energy.converters[ram_0.tag]                     )
    conditions.energy.converters[inlet_nozzle.tag]             = deepcopy(conditions.energy.converters[inlet_nozzle_0.tag]            ) 
    conditions.energy.converters[compressor.tag]               = deepcopy(conditions.energy.converters[compressor_0.tag] ) 
    conditions.energy.converters[combustor.tag]                = deepcopy(conditions.energy.converters[combustor_0.tag]               )
    conditions.energy.converters[low_pressure_turbine.tag]     = deepcopy(conditions.energy.converters[low_pressure_turbine_0.tag]    )
    conditions.energy.converters[high_pressure_turbine.tag]    = deepcopy(conditions.energy.converters[high_pressure_turbine_0.tag]   )
    conditions.energy.converters[core_nozzle.tag]              = deepcopy(conditions.energy.converters[core_nozzle_0.tag]             )

    # compute moment  
    moment_vector      = 0*state.ones_row(3)
    thrust_vector      = 0*state.ones_row(3)
    thrust_vector[:,0] = conditions.energy.propulsors[turboprop.tag].thrust[:,0] 
    moment_vector[:,0] = turboprop.origin[0][0] -   center_of_gravity[0][0] 
    moment_vector[:,1] = turboprop.origin[0][1]  -  center_of_gravity[0][1] 
    moment_vector[:,2] = turboprop.origin[0][2]  -  center_of_gravity[0][2]
    moment             = np.cross(moment_vector,thrust_vector)    

    power                                              = conditions.energy.propulsors[turboprop.tag].power 
    conditions.energy.propulsors[turboprop.tag].moment = moment
    
    power_elec = 0*state.ones_row(1)
    if compressor.motor != None and  len(state.numerics.time.differentiate) > 0: 
        conditions.energy.converters[compressor.motor.tag]  = deepcopy(conditions.energy.converters[compressor_0.motor.tag]) 
        power_elec =  conditions.energy.converters[compressor.motor.tag].outputs.power  
    
    if compressor.generator != None and len(state.numerics.time.differentiate) > 0:  
        conditions.energy.converters[compressor.generator.tag]  = deepcopy(conditions.energy.converters[compressor_0.generator.tag]) 
        power_elec =  conditions.energy.converters[compressor.generator.tag].inputs.power   

    return thrust_vector,moment,power, power_elec


# RCAIDE/Methods/Energy/Propulsors/Converters/turboshaft/__init__.py
# 

"""
Collection of methods for analyzing turboshaft propulsion systems. These methods handle the design, 
sizing, and performance analysis of turboshaft engines, including power output calculations, core sizing, 
and operational performance evaluation.

The module provides functions for:
    - Computing power output and fuel consumption
    - Sizing engine core components
    - Analyzing design point performance
    - Computing off-design performance characteristics
    - Managing stored engine performance data

See Also
--------
RCAIDE.Library.Components.Powertrain.Converters.Turboshaft_Propulsor
"""

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------

from .append_turboshaft_conditions   import append_turboshaft_conditions
from .compute_power                  import compute_power
from .size_core                      import size_core
from .design_turboshaft              import design_turboshaft
from .compute_turboshaft_performance import compute_turboshaft_performance , reuse_stored_turboshaft_data


# RCAIDE/Methods/Energy/Propulsors/Turbofan/compute_thrust.py
# 
# 
# Created:  Jul 2023, M. Clarke

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ---------------------------------------------------------------------------------------------------------------------- 
 # RCAIDE imports  
from RCAIDE.Framework.Core      import Units 

# Python package imports
import numpy as np
 
# ----------------------------------------------------------------------------------------------------------------------
#  compute_thrust
# ----------------------------------------------------------------------------------------------------------------------
def compute_thrust(turbofan,conditions):
    """
    Computes thrust and other performance metrics for a turbofan engine.
    
    Parameters
    ----------
    turbofan : RCAIDE.Library.Components.Propulsors.Turbofan
        Turbofan engine component with the following attributes:
            - tag : str
                Identifier for the turbofan
            - reference_temperature : float
                Reference temperature for mass flow scaling [K]
            - reference_pressure : float
                Reference pressure for mass flow scaling [Pa]
            - compressor_nondimensional_massflow : float
                Non-dimensional mass flow parameter [kg·√K/(s·Pa)]
            - SFC_adjustment : float
                Adjustment factor for specific fuel consumption
    conditions : RCAIDE.Framework.Mission.Common.Conditions
        Flight conditions with:
            - freestream : Data
                Freestream properties
                    - isentropic_expansion_factor : numpy.ndarray
                        Ratio of specific heats (gamma)
                    - velocity : numpy.ndarray
                        Freestream velocity [m/s]
                    - speed_of_sound : numpy.ndarray
                        Speed of sound [m/s]
                    - mach_number : numpy.ndarray
                        Freestream Mach number
                    - pressure : numpy.ndarray
                        Freestream pressure [Pa]
                    - gravity : numpy.ndarray
                        Gravitational acceleration [m/s²]
                    - specific_heat_at_constant_pressure : numpy.ndarray
                        Specific heat at constant pressure [J/(kg·K)]
            - energy.propulsors[turbofan.tag] : Data
                Turbofan-specific conditions
                    - fuel_to_air_ratio : numpy.ndarray
                        Fuel-to-air ratio
                    - total_temperature_reference : numpy.ndarray
                        Reference total temperature [K]
                    - total_pressure_reference : numpy.ndarray
                        Reference total pressure [Pa]
                    - fan_nozzle_exit_velocity : numpy.ndarray
                        Fan nozzle exit velocity [m/s]
                    - fan_nozzle_static_pressure : numpy.ndarray
                        Fan nozzle static pressure [Pa]
                    - fan_nozzle_area_ratio : numpy.ndarray
                        Fan nozzle area ratio
                    - core_nozzle_exit_velocity : numpy.ndarray
                        Core nozzle exit velocity [m/s]
                    - core_nozzle_static_pressure : numpy.ndarray
                        Core nozzle static pressure [Pa]
                    - core_nozzle_area_ratio : numpy.ndarray
                        Core nozzle area ratio
                    - flow_through_core : numpy.ndarray
                        Fraction of flow through core
                    - flow_through_fan : numpy.ndarray
                        Fraction of flow through fan
                    - bypass_ratio : numpy.ndarray
                        Bypass ratio
                    - throttle : numpy.ndarray
                        Throttle setting [0-1]
    
    Returns
    -------
    None
        Results are stored in conditions.energy.propulsors[turbofan.tag]:
            - thrust : numpy.ndarray
                Total thrust force [N]
            - fan_thrust : numpy.ndarray
                Thrust from fan stream [N]
            - core_thrust : numpy.ndarray
                Thrust from core stream [N]
            - thrust_specific_fuel_consumption : numpy.ndarray
                Thrust specific fuel consumption [kg/(N·hr)]
            - non_dimensional_thrust : numpy.ndarray
                Non-dimensional thrust
            - core_mass_flow_rate : numpy.ndarray
                Core mass flow rate [kg/s]
            - fuel_flow_rate : numpy.ndarray
                Fuel flow rate [kg/s]
            - power : numpy.ndarray
                Power output [W]
            - specific_impulse : numpy.ndarray
                Specific impulse [s]
    
    Notes
    -----
    This function implements a thermodynamic model for a turbofan engine to calculate
    thrust, fuel consumption, and other performance metrics. It computes thrust from
    both the core and fan streams, accounting for momentum and pressure forces.
    
    **Major Assumptions**
        * Perfect gas behavior
        * Thrust is calculated from momentum and pressure forces at the nozzle exits
    
    **Theory**
    The non-dimensional thrust components are calculated as:
    
    .. math::
        F_{nd,fan} = \\phi_{fan} \\cdot (\\gamma \\cdot M_0^2 \\cdot (V_{fan}/V_0 - 1) + A_{fan} \\cdot (P_{fan}/P_0 - 1))
        
        F_{nd,core} = \\phi_{core} \\cdot (\\gamma \\cdot M_0^2 \\cdot (V_{core}/V_0 - 1) + A_{core} \\cdot (P_{core}/P_0 - 1))
    
    where:
        * :math:`\\phi_{fan}` is the flow through fan fraction
        * :math:`\\phi_{core}` is the flow through core fraction
        * :math:`\\gamma` is the ratio of specific heats
        * :math:`M_0` is the freestream Mach number
        * :math:`V_{fan}` is the fan nozzle exit velocity
        * :math:`V_{core}` is the core nozzle exit velocity
        * :math:`V_0` is the freestream velocity
        * :math:`A_{fan}` is the fan nozzle area ratio
        * :math:`A_{core}` is the core nozzle area ratio
        * :math:`P_{fan}` is the fan nozzle static pressure
        * :math:`P_{core}` is the core nozzle static pressure
        * :math:`P_0` is the freestream pressure
    
    The specific thrust is then:
    
    .. math::
        F_{sp} = \\frac{F_{nd,fan} + F_{nd,core}}{\\gamma \\cdot M_0}
    
    References
    ----------
    [1] Cantwell, B., "AA283 Course Notes", Stanford University. https://web.stanford.edu/~cantwell/AA283_Course_Material/
    
    See Also
    --------
    RCAIDE.Library.Methods.Powertrain.Propulsors.Turbofan.compute_turbofan_performance
    RCAIDE.Library.Methods.Powertrain.Propulsors.Turbofan.size_core
    """      
    # Unpack flight conditions 
    gamma                       = conditions.freestream.isentropic_expansion_factor 
    u0                          = conditions.freestream.velocity
    a0                          = conditions.freestream.speed_of_sound
    M0                          = conditions.freestream.mach_number
    p0                          = conditions.freestream.pressure  
    g                           = conditions.freestream.gravity        

    # Unpack turbofan operating conditions and properties 
    Tref                        = turbofan.reference_temperature
    Pref                        = turbofan.reference_pressure
    mdhc                        = turbofan.compressor_nondimensional_massflow
    SFC_adjustment              = turbofan.specific_fuel_consumption_reduction_factor 
    turbofan_conditions         = conditions.energy.propulsors[turbofan.tag]
    f                           = turbofan_conditions.fuel_to_air_ratio
    total_temperature_reference = turbofan_conditions.total_temperature_reference
    total_pressure_reference    = turbofan_conditions.total_pressure_reference 
    flow_through_core           = turbofan_conditions.flow_through_core 
    flow_through_fan            = turbofan_conditions.flow_through_fan  
    V_fan_nozzle                = turbofan_conditions.fan_nozzle_exit_velocity
    fan_area_ratio              = turbofan_conditions.fan_nozzle_area_ratio
    P_fan_nozzle                = turbofan_conditions.fan_nozzle_static_pressure
    P_core_nozzle               = turbofan_conditions.core_nozzle_static_pressure
    V_core_nozzle               = turbofan_conditions.core_nozzle_exit_velocity
    core_area_ratio             = turbofan_conditions.core_nozzle_area_ratio                   
    bypass_ratio                = turbofan_conditions.bypass_ratio  

    # Compute  non dimensional thrust
    fan_thrust_nondim   = flow_through_fan*(gamma*M0*M0*(V_fan_nozzle/u0-1.) + fan_area_ratio*(P_fan_nozzle/p0-1.))
    core_thrust_nondim  = flow_through_core*(gamma*M0*M0*(V_core_nozzle/u0-1.) + core_area_ratio*(P_core_nozzle/p0-1.))

    thrust_nondim       = core_thrust_nondim + fan_thrust_nondim

    # Computing Specifc Thrust
    Fsp   = 1./(gamma*M0)*thrust_nondim
    Fsp_c = 1./(gamma*M0)*core_thrust_nondim
    Fsp_f = 1./(gamma*M0)*fan_thrust_nondim

    # Compute specific impulse
    Isp   = Fsp*a0*(1.+bypass_ratio)/(f*g)

    # Compute TSFC
    TSFC  = f*g/(Fsp*a0*(1.+bypass_ratio))*(1.-SFC_adjustment) * Units.hour # 1/s is converted to 1/hr here
 
    # Compute core mass flow
    mdot_core  = mdhc*np.sqrt(Tref/total_temperature_reference)*(total_pressure_reference/Pref)

    # Compute dimensional thrust
    FD2   = Fsp*a0*(1.+bypass_ratio)*mdot_core*turbofan_conditions.throttle
    FD2_f = Fsp_f*a0*(1.+bypass_ratio)*mdot_core*turbofan_conditions.throttle
    FD2_c = Fsp_c*a0*(1.+bypass_ratio)*mdot_core*turbofan_conditions.throttle

    # Compute power 
    power   = FD2*u0    

    # Compute fuel flow rate 
    fuel_flow_rate   = np.fmax(FD2*TSFC/g,np.array([0.]))*1./Units.hour

    # Pack turbofan outouts  
    turbofan_conditions.thrust                            = FD2 
    turbofan_conditions.fan_thrust                        = FD2_f 
    turbofan_conditions.core_thrust                       = FD2_c 
    turbofan_conditions.thrust_specific_fuel_consumption  = TSFC
    turbofan_conditions.non_dimensional_thrust            = Fsp  
    turbofan_conditions.power                             = power   
    turbofan_conditions.specific_impulse                  = Isp
    turbofan_conditions.core_mass_flow_rate               = mdot_core
    turbofan_conditions.fuel_flow_rate                    = fuel_flow_rate   
    
    return  


# RCAIDE/Methods/Powertrain/Sources/Fuel_Cells/Larminie_Model/compute_power.py
#  
# Created: Jan 2025, M. Clarke
# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------
from .compute_voltage import compute_voltage 
import numpy as np

# ----------------------------------------------------------------------------------------------------------------------
#  Find Power Larminie
# ----------------------------------------------------------------------------------------------------------------------
def compute_power(current_density, fuel_cell, sign=1.0):
    '''
    Function that determines the power output per cell, based on in 
    input current density
    
    Assumptions:
    None(calls other functions)
    
    Inputs:
    current_density      [Amps/m**2]
    fuel cell.
        interface area   [m**2]
        
    Outputs:
    power_out            [W]
    
    '''
    
    # sign variable is used so that you can maximize the power, by minimizing the -power
    i1            = current_density
    A             = fuel_cell.interface_area
    v             = compute_voltage(fuel_cell,current_density)   # useful voltage vector
    power_out     = sign* np.multiply(v,i1)*A                    # obtain power output in W/cell 
    return power_out


# RCAIDE/Library/Methods/Powertrain/Converters/Compression_Nozzle/compute_supersonic_nozzle_performance.py
# (c) Copyright 2023 Aerospace Research Community LLC
# 
# Created:  Jun 2024, M. Clarke 

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------     

# package imports
import numpy as np   
from RCAIDE.Library.Methods.Gas_Dynamics.fm_id import fm_id

# ---------------------------------------------------------------------------------------------------------------------- 
# compute_compression_nozzle_performance
# ----------------------------------------------------------------------------------------------------------------------    
def compute_supersonic_nozzle_performance(supersonic_nozzle,conditions): 
    """
    Computes the performance parameters of a supersonic nozzle based on input conditions.
    
    Parameters
    ----------
    supersonic_nozzle : RCAIDE.Components.Energy.Converters.Supersonic_Nozzle
        The supersonic nozzle component for which performance is being computed
    conditions : RCAIDE.Framework.Mission.Common.Conditions
        Container for flight conditions and energy system states
        
    Returns
    -------
    None
        This function modifies the conditions object in-place
    
    Notes
    -----
    This function calculates output values from input conditions according to 
    gas dynamics equations for supersonic nozzles.
    
    **Major Assumptions**
        * Constant polytropic efficiency and pressure ratio
        * Isentropic flow except for losses accounted by efficiency terms
        * Perfect gas behavior
    
    **Theory**
    
    The nozzle performance is calculated using isentropic flow relations. For supersonic flow,
    the Mach number is calculated from the pressure ratio:
    
    .. math::
        M = \\sqrt{\\frac{2}{\\gamma-1}\\left[\\left(\\frac{P_t}{P_0}\\right)^{\\frac{\\gamma-1}{\\gamma}}-1\\right]}
    
    The static temperature is related to stagnation temperature by:
    
    .. math::
        T = \\frac{T_t}{1+\\frac{\\gamma-1}{2}M^2}
    
    **Definitions**
    
    'Pressure ratio'
        Ratio of exit pressure to inlet pressure in the nozzle
    
    'Polytropic efficiency'
        Measure of the nozzle's thermodynamic efficiency accounting for irreversibilities
    
    References
    ----------
    [1] Cantwell, B. "AA283 Course Material: Course Notes." Stanford University. https://web.stanford.edu/~cantwell/AA283_Course_Material/AA283_Course_Notes/
    
    See Also
    --------
    RCAIDE.Library.Methods.Gas_Dynamics.fm_id
    """           
    
    #unpack the values
    
    #unpack from conditions
    gamma               = conditions.freestream.isentropic_expansion_factor
    Cp                  = conditions.freestream.specific_heat_at_constant_pressure
    Po                  = conditions.freestream.pressure
    Pto                 = conditions.freestream.stagnation_pressure
    Tto                 = conditions.freestream.stagnation_temperature
    R                   = conditions.freestream.gas_specific_constant
    Mo                  = conditions.freestream.mach_number
    s_nozzle_conditions = conditions.energy.converters[supersonic_nozzle.tag]
    
    #unpack from inputs
    Tt_in    = s_nozzle_conditions.inputs.stagnation_temperature
    Pt_in    = s_nozzle_conditions.inputs.stagnation_pressure
     
    pid      = supersonic_nozzle.pressure_ratio
    etapold  = supersonic_nozzle.polytropic_efficiency
    eta_rec  = supersonic_nozzle.pressure_recovery
    
    #Method for computing the nozzle properties
    
    #--Getting the output stagnation quantities
    Pt_out   = Pt_in*pid*eta_rec
    Tt_out   = Tt_in*(pid*eta_rec)**((gamma-1)/(gamma)*etapold)
    ht_out   = Cp*Tt_out
    
    
    #compute the output Mach number, static quantities and the output velocity
    Mach          = np.sqrt((((Pt_out/Po)**((gamma-1)/gamma))-1)*2/(gamma-1))
    
    #Remove check on mach numbers fromn expansion nozzle
    i_low         = Mach < 1.0
    
    #initializing the Pout array
    P_out         = 1.0 *Mach/Mach
    
    #Computing output pressure and Mach number for the case Mach <1.0
    P_out[i_low]  = Po[i_low]
    Mach[i_low]   = np.sqrt((((Pt_out[i_low]/Po[i_low])**((gamma[i_low]-1.)/gamma[i_low]))-1.)*2./(gamma[i_low]-1.))
    
    #Computing the output temperature,enthalpy, velocity and density
    T_out         = Tt_out/(1.+(gamma-1.)/2.*Mach*Mach)
    h_out         = Cp*T_out
    u_out         = np.sqrt(2.*(ht_out-h_out))
    rho_out       = P_out/(R*T_out)
    
    #Computing the freestream to nozzle area ratio (mainly from thrust computation)
    area_ratio    = (fm_id(Mo,gamma)/fm_id(Mach,gamma)*(1/(Pt_out/Pto))*(np.sqrt(Tt_out/Tto)))
    
    #pack computed quantities into outputs
    s_nozzle_conditions.outputs.stagnation_temperature  = Tt_out
    s_nozzle_conditions.outputs.stagnation_pressure     = Pt_out
    s_nozzle_conditions.outputs.stagnation_enthalpy     = ht_out
    s_nozzle_conditions.outputs.mach_number             = Mach
    s_nozzle_conditions.outputs.static_temperature      = T_out
    s_nozzle_conditions.outputs.density                 = rho_out
    s_nozzle_conditions.outputs.static_enthalpy         = h_out
    s_nozzle_conditions.outputs.velocity                = u_out
    s_nozzle_conditions.outputs.static_pressure         = P_out
    s_nozzle_conditions.outputs.area_ratio              = area_ratio
        
    return 


# RCAIDE/Library/Attributes/Propellants/Butaanol.py
#  
# Created:  Mar 2024, M. Clarke

# ---------------------------------------------------------------------------------------------------------------------- 
#  Imports
# ---------------------------------------------------------------------------------------------------------------------- 

from .Propellant import Propellant 

# ---------------------------------------------------------------------------------------------------------------------- 
#  Propanol Propellant Class
# ----------------------------------------------------------------------------------------------------------------------   
class Butanol(Propellant):
    """
    A class representing butanol (C4H9OH) fuel properties for propulsion applications.

    Attributes
    ----------
    tag : str
        Identifier for the propellant ('Butanol')
    reactant : str
        Oxidizer used for combustion ('O2')
    density : float
        Fuel density in kg/m³ (809.56)
    specific_energy : float
        Specific energy content in J/kg (3.61e7)
    energy_density : float
        Energy density in J/m³ (2.92e10)
    lower_heating_value : float
        Lower heating value in J/kg (3.44e7)
    use_high_fidelity_kinetics_model : bool
        Flag for using detailed chemical kinetics (False)
    fuel_surrogate_chemical_properties : dict
        Simplified chemical composition for surrogate model {'N1C4H9OH': 1.0}
    fuel_chemical_properties : dict
        Detailed chemical composition for high-fidelity model
    air_chemical_properties : dict
        Air composition for combustion calculations
    surrogate_species_list : list
        Species considered in surrogate model ['CO', 'CO2', 'H2O']
    species_list : list
        Species considered in detailed model ['CO', 'CO2', 'H2O', 'NO', 'NO2', 'CSOLID']
    global_warming_potential_100 : Data
        100-year global warming potential for emissions
            - CO2 : float
                GWP for carbon dioxide (1)
            - H2O : float
                GWP for water vapor (0.06)
            - SO2 : float
                GWP for sulfur dioxide (-226)
            - NOx : float
                GWP for nitrogen oxides (52)
            - Soot : float
                GWP for particulate matter (1166)
            - Contrails : float
                GWP for contrail formation (11)

    Notes
    -----
    This class implements properties for butanol fuel, including options forboth simplified and 
    detailed chemical kinetics. Properties are specified at standard conditions 
    (15°C, 1 atm).

    **Definitions**
    
    'Lower Heating Value'
        Heat of combustion excluding latent heat of water vapor
    
    'Global Warming Potential'
        Relative measure of heat trapped in atmosphere compared to CO2

    **Major Assumptions**
        * Properties are for standard temperature and pressure conditions (20C, 1atm)

    References
    ----------
    [1] IEA. (n.d.). Properties. AMF. https://www.iea-amf.org/content/fuel_information/butanol/properties#:~:text=Isobutanol%20or%20n%2Dbutanol%20are,from%20gasoline%20in%20normal%20conditions.&text=Viscosities%20of%20butanol%20isomers%20are,to%20viscosities%20of%20diesel%20fuel. 
    """

    def __defaults__(self):
        """This sets the default values.
    
        Assumptions:
            Density at 20C 1 atm
        
        Source: 
    
        """    
        self.tag                       = 'Butanol'
        self.reactant                  = 'O2'
        self.density                   = 809.56                            # kg/m^3 (15 C, 1 atm)
        self.specific_energy           = 3.61e7                            # J/kg
        self.energy_density            = 2.92e10                           # J/m^3
        self.lower_heating_value       = 3.44e7                            # J/kg

        self.stoichiometric_fuel_air_ratio = 0         # [-] Stoichiometric Fuel to Air ratio
        self.heat_of_vaporization          = 0         # [J/kg] Heat of vaporization at standard conditions
        self.temperature                   = 0         # [K] Temperature of fuel
        self.pressure                      = 0         # [Pa] Pressure of fuel
        self.fuel_surrogate_S1             = {} # [-] Mole fractions of fuel surrogate species
        self.kinetic_mechanism             = '' # [-] Kinetic mechanism for fuel surrogate species
        self.oxidizer                      = ''     
        
        self.global_warming_potential_100.CO2       = 1     # CO2e/kg  
        self.global_warming_potential_100.H2O       = 0.06  # CO2e/kg  
        self.global_warming_potential_100.SO2       = -226  # CO2e/kg  
        self.global_warming_potential_100.NOx       = 52    # CO2e/kg  
        self.global_warming_potential_100.Soot      = 1166  # CO2e/kg    
        self.global_warming_potential_100.Contrails = 11 # kg/CO2e/km          


# RCAIDE/Library/Methods/Weights/Correlation_Buildups/FLOPS/compute_landing_gear_weight.py
# 
# Created:  Sep 2024, M. Clarke
# Modified: Feb 2025, A. Molloy and S. Shekar

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------

# RCAIDE
import  RCAIDE
from RCAIDE.Framework.Core    import Units ,  Data

# python imports 
import  numpy as  np
 
# ----------------------------------------------------------------------------------------------------------------------
#  Landing Gear Weight 
# ----------------------------------------------------------------------------------------------------------------------
def compute_landing_gear_weight(vehicle):
    """
    Calculates the weight of a general aviation aircraft's landing gear system using FLOPS methodology.

    Parameters
    ----------
    vehicle : RCAIDE.Vehicle()
        Vehicle data structure containing:
            - mass_properties.max_takeoff
            - landing_gear.nose_wheels
            - landing_gear.main_wheels
            - landing_gear.configuration_type
            - flight_envelope.design_range
            - wings.main_wing.dihedral
            - fuselages.fuselage.width
            - fuselages.fuselage.lengths.total

    Returns
    -------
    output : Data()
        Data structure containing:
            - output.main: weight of the main landing gear in kilograms
            - output.nose: weight of the nose landing gear in kilograms

    Notes
    -----
    The function uses FLOPS (Flight Optimization System) statistical regression models 
    based on historical aircraft data to estimate landing gear weight. The calculation 
    includes the weight of struts, wheels, tires, brakes, and actuation systems. This method 
    is common between general aviation and transport aircraft in FLOPS. Of note, FLOPS does
    not use user defined landing gear lengths and tire sizes.
    
    **Major Assumptions**
        * Landing gear is not for a fighter jet, carrier based aircraft, or supersonic transport
        * Landing gear length is based on the wingspan and fuselage length
        * Brake system, retraction mechanism, and other components are baked into the weight
    
    **Theory**

    The FLOPS weight estimation follows the relationship:

    .. math::
        W_{LGM} = (0.0117) * (W_{LDG})^{0.95} * (X_{MLG})^{0.43}
        W_{LGN} = (0.048) * (W_{LDG})^{0.67} * (X_{NLG})^{0.43}
    Where:
        - W_{LGM} is landing gear weight for the main landing gear
        - W_{LGN} is landing gear weight for the nose landing gear
        - W_{LDG} is landing weight for the entire aircraft
        - X_{MLG} is the length of the main landing gear
        - X_{NLG} is the length of the nose landing gear
    
    Reference:
    """
    RFACT = 0.00004 # Not a supersonic transport aircraft
    
    DESRNG  = vehicle.flight_envelope.design_range / Units.nmi  # Design range in nautical miles
    WLDG    = vehicle.mass_properties.max_takeoff / Units.lbs * (1 - RFACT * DESRNG)
    
    for wing in vehicle.wings:
        if isinstance(wing,RCAIDE.Library.Components.Wings.Main_Wing):
            main_wing = wing
    
    l_f =  0
    for fuselage in vehicle.fuselages:
        if l_f < fuselage.lengths.total:
            main_fuselage = fuselage 
            l_f = main_fuselage.lengths.total
        
    for network in vehicle.networks:
        for propulsor in  network.propulsors:
            if propulsor.wing_mounted:
                FNAC = 0
                if isinstance(propulsor, RCAIDE.Library.Components.Powertrain.Propulsors.Turbofan) or  isinstance(propulsor, RCAIDE.Library.Components.Powertrain.Propulsors.Turbojet):
                    nacelle =  propulsor.nacelle 
                    FNAC    = nacelle.diameter / Units.ft
                if isinstance(propulsor, RCAIDE.Library.Components.Powertrain.Propulsors.Internal_Combustion_Engine) or isinstance(propulsor, RCAIDE.Library.Components.Powertrain.Propulsors.Constant_Speed_Internal_Combustion_Engine): 
                    FNAC    = propulsor.propeller.tip_radius * 2 / Units.ft
                if isinstance(propulsor, RCAIDE.Library.Components.Powertrain.Propulsors.Electric_Ducted_Fan): 
                    FNAC    = propulsor.ducted_fan.tip_radius * 2 / Units.ft                          
                if isinstance(propulsor, RCAIDE.Library.Components.Powertrain.Propulsors.Electric_Rotor): 
                    FNAC    = propulsor.rotor.tip_radius * 2 / Units.ft                          
                DIH     = main_wing.dihedral
                YEE     = np.max(np.abs(np.array(propulsor.origin)[:, 1])) / Units.inches
                WF      = main_fuselage.width / Units.ft
                XMLG    = 12 * FNAC + (0.26 - np.tan(DIH)) * (YEE - 6 * WF)  # length of extended main landing gear
            else:
                XMLG    = 0.75 * main_fuselage.lengths.total / Units.ft  # length of extended nose landing gear
    XNLG = 0.7 * XMLG
    WLGM = (0.0117) * WLDG ** 0.95 * XMLG ** 0.43
    WLGN = (0.048) * WLDG ** 0.67 * XNLG ** 0.43

    output      = Data()
    output.main = WLGM * Units.lbs
    output.nose = WLGN * Units.lbs
    return output


# RCAIDE/Library/Methods/Powertrain/Propulsors/Constant_Speed_ICE_Propulsor/append_constant_speed_internal_combustion_engine_conditions.py
# 
# Created:  Jun 2024, M. Clarke

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------
# RCAIDE imports
import RCAIDE
from RCAIDE.Framework.Mission.Common     import   Conditions

# ---------------------------------------------------------------------------------------------------------------------- 
#  append_constant_speed_internal_combustion_engine_conditions
# ----------------------------------------------------------------------------------------------------------------------    
def append_constant_speed_internal_combustion_engine_conditions(propulsor, segment, energy_conditions, noise_conditions):
    """
    Initializes constant speed internal combustion engine operating conditions for a mission segment.
    
    Parameters
    ----------
    propulsor : RCAIDE.Library.Components.Propulsors.Constant_Speed_Internal_Combustion_Engine
        Constant speed internal combustion engine propulsor component with the following attributes:
            - tag : str
                Identifier for the engine
            - items : dict
                Dictionary of subcomponents
    segment : RCAIDE.Framework.Mission.Segments.Segment
        Mission segment with the following attributes:
            - state : Data
                Segment state
                    - ones_row : function
                        Function to create array of ones with specified length
    energy_conditions : RCAIDE.Framework.Mission.Common.Conditions
        Energy conditions container where engine conditions will be stored
    noise_conditions : RCAIDE.Framework.Mission.Common.Conditions
        Noise conditions container where engine noise conditions will be stored
    
    Returns
    -------
    None
        Results are stored in energy_conditions.propulsors[propulsor.tag] and
        segment.state.conditions.noise[propulsor.tag]
    
    Notes
    -----
    This function initializes the necessary data structures for storing constant speed
    internal combustion engine operating conditions during a mission segment. It creates 
    zero-filled arrays for various performance parameters and recursively calls the 
    append_operating_conditions method for each subcomponent of the engine.
    
    The function initializes the following parameters in energy_conditions:
        - throttle
        - commanded_thrust_vector_angle
        - thrust
        - power
        - moment
        - fuel_flow_rate
        - inputs and outputs containers
    
    It also creates a noise conditions container for the engine.
    
    See Also
    --------
    RCAIDE.Library.Methods.Powertrain.Propulsors.Constant_Speed_Internal_Combustion_Engine.compute_constant_speed_internal_combustion_engine_performance
    """
    # unpack 
    ones_row          = segment.state.ones_row 
    
    # add propulsor conditions 
    energy_conditions.propulsors[propulsor.tag]                               = Conditions()  
    energy_conditions.propulsors[propulsor.tag].throttle                      = 0. * ones_row(1)      
    energy_conditions.propulsors[propulsor.tag].commanded_thrust_vector_angle = 0. * ones_row(1)  
    energy_conditions.propulsors[propulsor.tag].thrust                        = 0. * ones_row(3) 
    energy_conditions.propulsors[propulsor.tag].power                         = 0. * ones_row(1) 
    energy_conditions.propulsors[propulsor.tag].moment                        = 0. * ones_row(3) 
    energy_conditions.propulsors[propulsor.tag].fuel_flow_rate                = 0. * ones_row(1)      
    energy_conditions.propulsors[propulsor.tag].inputs                        = Conditions()
    energy_conditions.propulsors[propulsor.tag].outputs                       = Conditions() 
    segment.state.conditions.noise[propulsor.tag]                             = Conditions()
 
    # parse propulsor for comoonent and append 
    for tag, item in  propulsor.items(): 
        if issubclass(type(item), RCAIDE.Library.Components.Component):
            item.append_operating_conditions(segment,energy_conditions,noise_conditions=noise_conditions) 
            for sub_tag, sub_item in  item.items(): 
                if issubclass(type(sub_item), RCAIDE.Library.Components.Component): 
                    sub_item.append_operating_conditions(segment,energy_conditions)    
    return 


# RCAIDE/Library/Methods/Powertrain/Converters/Ram/append_ram_conditions.py
# 
# Created:  Jun 2024, M. Clarke  

from RCAIDE.Framework.Mission.Common     import   Conditions

# ---------------------------------------------------------------------------------------------------------------------- 
#  append_ram_conditions
# ----------------------------------------------------------------------------------------------------------------------    
def append_ram_conditions(ram, segment, energy_conditions):
    """
    Initializes ram air converter operating conditions for a mission segment.
    
    Parameters
    ----------
    ram : RCAIDE.Library.Components.Converters.Ram
        Ram air converter component with the following attributes:
            - tag : str
                Identifier for the ram air converter
    segment : RCAIDE.Framework.Mission.Segments.Segment
        Mission segment with the following attributes:
            - state : Data
                Segment state
                    - ones_row : function
                        Function to create array of ones with specified length
    energy_conditions : RCAIDE.Framework.Mission.Common.Conditions
        Energy conditions container where ram air converter conditions will be stored
    
    Returns
    -------
    None
    
    Notes
    -----
    This function initializes the necessary data structures for storing ram air converter
    operating conditions during a mission segment. It creates empty containers for
    input and output conditions that will be populated during the mission analysis.
    
    The function initializes the following in energy_conditions.converters[ram.tag]:
        - inputs : Conditions
            Input conditions container (empty)
        - outputs : Conditions
            Output conditions container (empty)
    
    The ram air converter is a component that captures the energy of the incoming airflow
    and converts it to a form usable by the propulsion system. It typically represents
    the inlet of a gas turbine engine or other air-breathing propulsion system.
    
    See Also
    --------
    RCAIDE.Library.Methods.Powertrain.Converters.Ram.compute_ram_performance
    """
    energy_conditions.converters[ram.tag]                              = Conditions() 
    energy_conditions.converters[ram.tag].inputs                       = Conditions() 
    energy_conditions.converters[ram.tag].outputs                      = Conditions() 
    return 


# RCAIDE/Library/Methods/Powertrain/Converters/Generator/append_generator_conditions.py
# 
# Created:  Feb 2025, M. Guidotti 
from RCAIDE.Framework.Mission.Common     import   Conditions

# ---------------------------------------------------------------------------------------------------------------------- 
#  append_generator_conditions
# ----------------------------------------------------------------------------------------------------------------------    
def append_generator_conditions(generator, segment, conditions):
    """
    Initializes generator operating conditions for a mission segment.
    
    Parameters
    ----------
    generator : RCAIDE.Library.Components.Converters.Generator
        Generator component with the following attributes:
            - tag : str
                Identifier for the generator
    segment : RCAIDE.Framework.Mission.Segments.Segment
        Mission segment with the following attributes:
            - state : Data
                Segment state
                    - ones_row : function
                        Function to create array of ones with specified length
    conditions : RCAIDE.Framework.Mission.Common.Conditions
        Energy conditions container where generator conditions will be stored
    
    Returns
    -------
    None
        Results are stored in conditions.converters[generator.tag]
    
    Notes
    -----
    This function initializes the necessary data structures for storing generator
    operating conditions during a mission segment. It creates zero-filled arrays
    for various input and output parameters.
    
    The function initializes the following in conditions.converters[generator.tag]:
        - inputs : Conditions
            Input conditions container
                - torque : numpy.ndarray
                    Input torque [N·m], initialized with zeros
                - power : numpy.ndarray
                    Input mechanical power [W], initialized with zeros
                - omega : numpy.ndarray
                    Angular velocity [rad/s], initialized with zeros
        - outputs : Conditions
            Output conditions container
                - current : numpy.ndarray
                    Output current [A], initialized with zeros
                - voltage : numpy.ndarray
                    Output voltage [V], initialized with zeros
        
    These parameters will be populated during the mission analysis to track the
    generator's performance as it converts mechanical power to electrical power.
    
    See Also
    --------
    RCAIDE.Library.Methods.Powertrain.Converters.Generator.compute_generator_performance
    """
    ones_row                                             = segment.state.ones_row                  
    conditions.converters[generator.tag]                 = Conditions()
    conditions.converters[generator.tag].inputs          = Conditions()
    conditions.converters[generator.tag].inputs.torque   = 0. * ones_row(1) 
    conditions.converters[generator.tag].inputs.power    = 0. * ones_row(1)
    conditions.converters[generator.tag].inputs.omega    = 0. * ones_row(1)
    conditions.converters[generator.tag].outputs         = Conditions()
    conditions.converters[generator.tag].outputs.current = 0. * ones_row(1) 
    conditions.converters[generator.tag].outputs.voltage = 0. * ones_row(1) 
    
    return 




# RCAIDE/Library/__init__.py 

"""
RCAIDE Library provides core functionality for aircraft design and optimization through a collection of modules.
The module contains the methods and classes used for modeling aircraft components and systems. The 
Framework module contains the system architecture that calls to different library modules. The 
modules work together to enable aircraft conceptual design, analysis, and optimization workflows.

See Also
--------
RCAIDE.Framework : Core framework functionality and tools
"""

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------
 
from . import Attributes
from . import Components 
from . import Methods 
from . import Mission
from . import Plots 


# RCAIDE/Methods/Library/Methods/Powertrain/Converters/compute_turboelectric_generator_performance.py
# 
# 
# Created:  Jul 2023, M. Clarke

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------
# RCAIDE imports      
from RCAIDE.Framework.Core import Data    
from RCAIDE.Library.Methods.Powertrain.Converters.Turboshaft         import compute_turboshaft_performance
from RCAIDE.Library.Methods.Powertrain.Converters.Generator          import compute_generator_performance 
 
# python imports 
from copy import deepcopy 
import numpy as np
# ----------------------------------------------------------------------------------------------------------------------
# compute_turboelectric_generator_performance
# ---------------------------------------------------------------------------------------------------------------------- 
def compute_turboelectric_generator_performance(turboelectric_generator, state, fuel_line=None, bus=None):
    """
    Computes the performance of a turboelectric generator system.
    
    Parameters
    ----------
    turboelectric_generator : RCAIDE.Components.Energy.Converters.Turboelectric_Generator
        The turboelectric generator component for which performance is being computed
    state : RCAIDE.Framework.Mission.Common.State
        Container for mission segment conditions
    fuel_line : RCAIDE.Components.Energy.Distribution.Fuel_Line, optional
        Fuel distribution system connected to the turboelectric generator
    bus : RCAIDE.Components.Energy.Distribution.Bus, optional
        Electrical bus connected to the generator output
        
    Returns
    -------
    P_mech : float
        Mechanical power produced by the turboshaft engine [W]
    P_elec : float
        Electrical power produced by the generator [W]
    stored_results_flag : bool
        Flag indicating that results have been stored for potential reuse
    stored_propulsor_tag : str
        Tag identifier of the turboelectric generator with stored results
        
    Notes
    -----
    This function handles both direct and inverse calculations for the turboelectric generator:
        - Direct calculation (inverse_calculation=False): Computes generator output based on 
        turboshaft throttle setting
        - Inverse calculation (inverse_calculation=True): Determines turboshaft fuel consumption 
        based on required generator output power
    
    The function coordinates the operation of the turboshaft engine and generator components,
    ensuring proper power flow and electrical characteristics.
    
    **Major Assumptions**
        * The turboshaft and generator are properly connected and compatible
        * Bus voltage is constant across all operating conditions
        * Mechanical power from turboshaft is directly coupled to generator input
    
    See Also
    --------
    RCAIDE.Library.Methods.Powertrain.Converters.Turboshaft.compute_turboshaft_performance
    RCAIDE.Library.Methods.Powertrain.Converters.Generator.compute_generator_performance
    """

    conditions                         = state.conditions
    generator                          = turboelectric_generator.generator
    turboshaft                         = turboelectric_generator.turboshaft 
    compressor                         = turboshaft.compressor
    turboelectric_generator_conditions = conditions.energy.converters[turboelectric_generator.tag] 
    generator_conditions               = conditions.energy.converters[generator.tag]
    turboshaft_conditions              = conditions.energy.converters[turboshaft.tag]
    compressor_conditions              = conditions.energy.converters[compressor.tag]
    
    generator.inverse_calculation      =  turboelectric_generator.inverse_calculation
    turboshaft.inverse_calculation     =  turboelectric_generator.inverse_calculation
    
    if turboelectric_generator.inverse_calculation == False:
        # here we run the turboshaft first, then run the generator
        turboshaft_conditions.throttle = turboelectric_generator_conditions.throttle
        
        # run the generator 
        P_mech,stored_results_flag,stored_propulsor_tag = compute_turboshaft_performance(turboshaft,state,turboelectric_generator,fuel_line)
        
        # connect properties of the turboshaft to generator 
        generator_conditions.inputs.power  = P_mech     
        generator_conditions.inputs.omega  = compressor_conditions.omega         
        
        # assign voltage across bus 
        generator_conditions.outputs.voltage = bus.voltage*np.ones_like(generator_conditions.inputs.power)
        
         # run the generator 
        compute_generator_performance(generator,conditions)   
        turboelectric_generator_conditions.fuel_flow_rate =  turboshaft_conditions.fuel_flow_rate  
         
    else:
        # here , we know the electric power produced by the generator and we want to determine how much fuel was used to produce said power
        
        # assign voltage across bus 
        generator_conditions.outputs.voltage = bus.voltage*np.ones_like(generator_conditions.outputs.power)
        
        # run the generator 
        compute_generator_performance(generator,conditions)
        
        # connect properties of the generator to the turboshaft 
        turboshaft_conditions.power  = generator_conditions.inputs.power
        
        # run the turboshaft 
        P_mech,stored_results_flag,stored_propulsor_tag = compute_turboshaft_performance(turboshaft,state,turboelectric_generator,fuel_line) 
        turboelectric_generator_conditions.fuel_flow_rate =  turboshaft_conditions.fuel_flow_rate   
    
    P_elec                      = generator_conditions.outputs.power       
    
    # Pack results      
    stored_results_flag    = True
    stored_propulsor_tag   = turboelectric_generator.tag
    
    return P_mech,P_elec,stored_results_flag,stored_propulsor_tag

def reuse_stored_turboelectric_generator_data(turboelectric_generator,state,fuel_line,bus,stored_converter_tag,center_of_gravity= [[0.0, 0.0,0.0]]):
    '''Reuses results from one turboelectric_generator for identical propulsors
    
    Assumptions: 
    N/A

    Source:
    N/A

    Inputs:  
    conditions           - operating conditions data structure     [-]  
    fuel_line            - fuelline                                [-] 
    turboelectric_generator           - turboelectric_generator data structure              [-] 
    total_power          - power of turboelectric_generator group               [W] 

    Outputs:  
    total_power          - power of turboelectric_generator group               [W] 
    
    Properties Used: 
    N.A.        
    '''
 
    conditions                  = state.conditions 
    generator                   = turboelectric_generator.generator
    turboshaft                  = turboelectric_generator.turboshaft
    ram                         = turboelectric_generator.ram
    inlet_nozzle                = turboelectric_generator.inlet_nozzle 
    compressor                  = turboelectric_generator.compressor
    high_pressure_compressor    = turboelectric_generator.high_pressure_compressor
    combustor                   = turboelectric_generator.combustor 
    low_pressure_turbine        = turboelectric_generator.low_pressure_turbine
    core_nozzle                 = turboelectric_generator.core_nozzle
    generator_0                 = fuel_line.converters[stored_converter_tag].generator 
    turboshaft_0                = fuel_line.converters[stored_converter_tag].turboshaft
    ram_0                       = fuel_line.converters[stored_converter_tag].ram
    inlet_nozzle_0              = fuel_line.converters[stored_converter_tag].inlet_nozzle 
    compressor_0                = fuel_line.converters[stored_converter_tag].compressor
    high_pressure_compressor_0  = fuel_line.converters[stored_converter_tag].high_pressure_compressor
    combustor_0                 = fuel_line.converters[stored_converter_tag].combustor
    low_pressure_turbine_0      = fuel_line.converters[stored_converter_tag].low_pressure_turbine
    core_nozzle_0               = fuel_line.converters[stored_converter_tag].core_nozzle

    # deep copy results 
    conditions.energy.converters[generator.tag]                = deepcopy(conditions.energy.converters[generator_0.tag]              ) 
    conditions.energy.converters[turboshaft.tag]               = deepcopy(conditions.energy.converters[turboshaft_0.tag]             )
    conditions.energy.converters[turboelectric_generator.tag]  = deepcopy(conditions.energy.converters[stored_converter_tag]) 
    conditions.energy.converters[ram.tag]                      = deepcopy(conditions.energy.converters[ram_0.tag]                     )
    conditions.energy.converters[inlet_nozzle.tag]             = deepcopy(conditions.energy.converters[inlet_nozzle_0.tag]            ) 
    conditions.energy.converters[compressor.tag]               = deepcopy(conditions.energy.converters[compressor_0.tag] )
    conditions.energy.converters[high_pressure_compressor.tag] = deepcopy(conditions.energy.converters[high_pressure_compressor_0.tag])
    conditions.energy.converters[combustor.tag]                = deepcopy(conditions.energy.converters[combustor_0.tag]               )
    conditions.energy.converters[low_pressure_turbine.tag]     = deepcopy(conditions.energy.converters[low_pressure_turbine_0.tag]    ) 
    conditions.energy.converters[core_nozzle.tag]              = deepcopy(conditions.energy.converters[core_nozzle_0.tag]             ) 
 
    P_elec         = conditions.energy.converters[generator.tag].outputs.power 
    P_mech         = conditions.energy.converters[turboshaft.tag].outputs.power  
    
    return P_mech, P_elec
 


# RCAIDE/Methods/Powertrain/Sources/Cryogenic_Tanks/append_cryogenic_tank_conditions.py
# 
# 
# Created:  Jan 2025, M. Clarke

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------
# RCAIDE imports  
from RCAIDE.Framework.Mission.Common     import   Conditions

# ----------------------------------------------------------------------------------------------------------------------
#  METHOD
# ----------------------------------------------------------------------------------------------------------------------  
def append_cryogenic_tank_conditions(cryogenic_tank, segment, bus):
    """
    Appends conditions data structure to cryogenic tank component for use during mission analysis.
    
    Parameters
    ----------
    cryogenic_tank : CryogenicTank
        The cryogenic tank component for which conditions are being initialized.
    segment : Segment
        The mission segment in which the cryogenic tank is operating.
    bus : Bus
        The energy bus connected to the cryogenic tank.
    
    Returns
    -------
    None
    
    Notes
    -----
    This function initializes the conditions for a cryogenic tank component at the start
    of a mission segment. It creates a Conditions object for the cryogenic tank within
    the segment's energy conditions dictionary, indexed by the bus tag and tank tag.
    
    The function initializes arrays for:
        - Mass flow rate [kg/s]
        - Mass [kg]
    
    These arrays are initialized with ones of the same length as the segment's state vector,
    which will be updated during mission analysis based on the cryogenic tank's performance
    and usage.
    
    See Also
    --------
    RCAIDE.Library.Components.Powertrain.Sources.Cryogenic_Tanks
    RCAIDE.Library.Methods.Powertrain.Sources.Fuel_Tanks.append_fuel_tank_conditions
    """
    ones_row    = segment.state.ones_row                 
    segment.state.conditions.energy[bus.tag][cryogenic_tank.tag]                 = Conditions()  
    segment.state.conditions.energy[bus.tag][cryogenic_tank.tag].mass_flow_rate  = ones_row(1)  
    segment.state.conditions.energy[bus.tag][cryogenic_tank.tag].mass            = ones_row(1)
    
    return 



# RCAIDE/Library/Missions/Common/Initialize/weights.py
# 
# 
# Created:  Jul 2023, M. Clarke
 

# ----------------------------------------------------------------------------------------------------------------------
#  Initialize Weights
# ---------------------------------------------------------------------------------------------------------------------- 
def weights(segment):
    """
    Initializes vehicle mass properties for mission segment analysis

    Parameters
    ----------
    segment : Segment
        The mission segment being analyzed

    Notes
    -----
    This function sets up the initial mass values for the vehicle. It determines
    the initial mass through a priority system and maintains mass continuity
    between segments.

    The function follows this priority for mass initialization:
        1. Previous segment final mass (if initials exist)
        2. Vehicle takeoff mass (if weight analysis exists)
        3. Network mass properties (fallback option)

    **Required Segment State Variables**

    If segment.state.initials exists:
        state.initials.conditions.weights:
            - total_mass : array
                Previous segment final mass [kg]

    state.conditions.weights:
        - total_mass : array
            Current segment mass array [kg]

    **Required Analysis Components**
    
    Either:
    segment.analyses.weights:
        - vehicle.mass_properties.takeoff : float
            Vehicle takeoff mass [kg]
    Or:
    segment.analyses.energy.vehicle.networks:
        - mass_properties.mass : float
            Network mass properties [kg]

    **Major Assumptions**
        * Continuous mass tracking when using initials
        * Valid mass values (positive)
        * At least one mass property source available
        * Mass measured in kilograms

    Returns
    -------
    None
        Updates segment conditions directly

    See Also
    --------
    RCAIDE.Framework.Mission.Segments
    """      
 
    if segment.state.initials:
        m_initial = segment.state.initials.conditions.weights.total_mass[-1,0] 
    else: 
        if segment.analyses.weights != None: 
            m_initial = segment.analyses.weights.vehicle.mass_properties.takeoff
        else: 
            m_initial = segment.analyses.energy.vehicle.mass_properties.takeoff

    m_current = segment.state.conditions.weights.total_mass
    
    segment.state.conditions.weights.total_mass[:,:] = m_current + (m_initial - m_current[0,0])
        
    return 


# RCAIDE/Library/Methods/Weights/Correlation_Buildups/FLOPS/compute_vertical_tail_weight.py
# 
# 
# Created:  Sep 2024, M. Clarke

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------

# RCAIDE 
from RCAIDE.Framework.Core    import Units 
 
# ----------------------------------------------------------------------------------------------------------------------
#  Vertical Tail Weight 
# ----------------------------------------------------------------------------------------------------------------------
def compute_vertical_tail_weight(vehicle, wing):
    """
    Computes the vertical tail weight using NASA FLOPS weight estimation method.

    Parameters
    ----------
    vehicle : Vehicle
        The vehicle instance containing:
            - mass_properties.max_takeoff : float
                Maximum takeoff weight [kg]
    wing : Wing
        The vertical tail instance containing:
            - taper : float
                Taper ratio
            - areas.reference : float
                Planform area [m²]

    Returns
    -------
    WVT : float
        Vertical tail weight [kg]

    Notes
    -----
    Uses FLOPS correlation developed from transport aircraft database. The method accounts 
    for tail size, aircraft weight, and planform effects.

    **Major Assumptions**
        * Conventional tail configuration
        * Single vertical tail (NVERT = 1). Multiple vertical tails are computed separately and summed.

    **Theory**
    Weight is computed using:
    .. math::
        W_{VT} = 0.32W_{TO}^{0.3}(\\lambda_{VT} + 0.5)N_{VT}^{0.7}S_{VT}^{0.85}

    where:
        * W_TO = takeoff gross weight
        * λ_VT = vertical tail taper ratio
        * N_VT = number of vertical tails (fixed at 1)
        * S_VT = vertical tail area

    The correlation accounts for:
        * Bending and torsional stiffness requirements
        * Rudder integration
        * Attachment structure
        * Standard design margins
        * Directional stability requirements

    References
    ----------
    [1] NASA Flight Optimization System (FLOPS)

    See Also
    --------
    RCAIDE.Library.Methods.Weights.Correlation_Buildups.FLOPS.compute_horizontal_tail_weight
    """
    DG          = vehicle.mass_properties.max_takeoff / Units.lbs  # Design gross weight in lb
    TRVT        = wing.taper
    NVERT       = 1  # Number of vertical tails
    WVT         = 0.32 * DG ** 0.3 * (TRVT + 0.5) * NVERT ** 0.7 * (wing.areas.reference/Units.ft**2)**0.85
    return WVT * Units.lbs





# RCAIDE/Library/Attributes/Propellants/Methane.py
#  
# Created:  Mar 2024, M. Clarke

# ---------------------------------------------------------------------------------------------------------------------- 
#  Imports
# ---------------------------------------------------------------------------------------------------------------------- 

from .Propellant import Propellant 

# ---------------------------------------------------------------------------------------------------------------------- 
# Methane Propellant Class
# ----------------------------------------------------------------------------------------------------------------------  
class Methane(Propellant):
    """
    A class representing methane (CH4) fuel properties and emissions characteristics 
    for propulsion applications.

    Attributes
    ----------
    tag : str
        Identifier for the propellant ('Methane')
    reactant : str
        Oxidizer used for combustion ('O2')
    density : float
        Fuel density in kg/m³ (422.6)
    specific_energy : float
        Specific energy content in J/kg (5.34e7)
    energy_density : float
        Energy density in J/m³ (2.26e10)
    lower_heating_value : float
        Lower heating value in J/kg (5.0e7)
    use_high_fidelity_kinetics_model : bool
        Flag for using detailed chemical kinetics (False)
    fuel_surrogate_chemical_properties : dict
        Simplified chemical composition {'CH4': 1.0}
    fuel_chemical_properties : dict
        Detailed chemical composition {'CH4': 1.0}
    air_chemical_properties : dict
        Air composition for combustion calculations
        {'O2': 0.2095, 'N2': 0.7809, 'AR': 0.0096}
    global_warming_potential_100 : Data
        100-year global warming potentials
            - CO2 : float
                Carbon dioxide (1)
            - H2O : float
                Water vapor (0.06)
            - SO2 : float
                Sulfur dioxide (-226)
            - NOx : float
                Nitrogen oxides (52)
            - Soot : float
                Particulate matter (1166)
            - Contrails : float
                Contrail formation (11)

    Notes
    -----
    Methane is the simplest hydrocarbon fuel and main component of natural gas. 
    It offers reduced carbon emissions compared to conventional fuels due to its 
    high hydrogen-to-carbon ratio.

    **Definitions**
    
    'Lower Heating Value'
        Heat of combustion excluding latent heat of water vapor
    
    'Global Warming Potential'
        Relative measure of heat trapped in atmosphere compared to CO2
    
    'Specific Energy'
        Energy content per unit mass, higher than conventional jet fuels

    **Major Assumptions**
        * Properties are for standard temperature and pressure conditions (15C, 1atm)
        * Pure methane composition (no higher hydrocarbons)
        * Standard atmospheric composition for air 
    """

    def __defaults__(self):
        """This sets the default values.
    
        Assumptions:
            Density at -162C, 1 atm
        
        Source:  
        """    
        self.tag                       = 'Methane'
        self.reactant                  = 'O2'
        self.density                   = 422.6                            # kg/m^3 (15 C, 1 atm)
        self.specific_energy           = 5.34e7                           # J/kg
        self.energy_density            = 2.26e10                          # J/m^3
        self.lower_heating_value       = 5.0e7                            # J/kg  
        
        self.stoichiometric_fuel_air_ratio = 0         # [-] Stoichiometric Fuel to Air ratio
        self.heat_of_vaporization          = 0         # [J/kg] Heat of vaporization at standard conditions
        self.temperature                   = 0         # [K] Temperature of fuel
        self.pressure                      = 0         # [Pa] Pressure of fuel
        self.fuel_surrogate_S1             = {} # [-] Mole fractions of fuel surrogate species
        self.kinetic_mechanism             = '' # [-] Kinetic mechanism for fuel surrogate species
        self.oxidizer                      = ''      
        
        self.global_warming_potential_100.CO2       = 1     # CO2e/kg  
        self.global_warming_potential_100.H2O       = 0.06  # CO2e/kg  
        self.global_warming_potential_100.SO2       = -226  # CO2e/kg  
        self.global_warming_potential_100.NOx       = 52    # CO2e/kg  
        self.global_warming_potential_100.Soot      = 1166  # CO2e/kg    
        self.global_warming_potential_100.Contrails = 11    # kg/CO2e/km          


# RCAIDE/Methods/Energy/Propulsors/Rotor_Design/optimization_setup.py
# 
# 
# Created:  Jul 2023, M. Clarke 

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------  
 
# RCAIDE Imports  
import RCAIDE 
from RCAIDE.Framework.Core                                                                    import Units, Data   
from RCAIDE.Framework.Optimization.Common                                                     import Nexus       
from RCAIDE.Library.Methods.Powertrain.Converters.Rotor.Design.blade_geometry_setup    import blade_geometry_setup
from RCAIDE.Library.Methods.Powertrain.Converters.Rotor.Design.procedure_setup         import procedure_setup

# Python package imports   
import numpy as np  
    
# ----------------------------------------------------------------------------------------------------------------------  
#  Optimization Setuo 
# ----------------------------------------------------------------------------------------------------------------------    
def optimization_setup(rotor, number_of_stations, print_iterations):
    """
    Sets up rotor optimization problem including design variables, constraints and objective function
    using RCAIDE's Nexus optimization framework.
    
    Parameters
    ----------
    rotor : RCAIDE.Library.Components.Powertrain.Converters.Rotor
        Rotor component (Lift_Rotor or Prop_Rotor) with optimization parameters
    number_of_stations : int
        Number of radial stations for blade discretization
    print_iterations : bool
        Flag to print optimization iterations
    
    Returns
    -------
    nexus : RCAIDE.Framework.Optimization.Common.Nexus
        RCAIDE's optimization framework object with the following attributes:
            - optimization_problem : Data
                Optimization problem definition
                    - inputs : numpy.ndarray
                        Design variables
                    - objective : numpy.ndarray
                        Objective function
                    - constraints : numpy.ndarray
                        Constraints
            - vehicle_configurations : list
                List of vehicle configurations for analysis
            - procedure : Data
                Optimization procedure
            - print_iterations : bool
                Flag to print optimization iterations
    
    Notes
    -----
    This function configures a complete optimization problem for rotor blade design by:
        1. Creating a Nexus optimization framework
        2. Validating the rotor type (must be Lift_Rotor or Prop_Rotor)
        3. Setting up design variables with bounds and scaling
        4. Defining the objective function
        5. Establishing constraints for performance and geometry
        6. Creating aliases to link optimization variables to vehicle properties
        7. Setting up vehicle configurations using blade_geometry_setup
        8. Configuring the optimization procedure
    
    The design variables include:
        - chord_r, chord_p, chord_q, chord_t: Parameters defining the chord distribution
        - twist_r, twist_p, twist_q, twist_t: Parameters defining the twist distribution
        - hover_tip_mach: Tip Mach number in hover
        - OEI_tip_mach: Tip Mach number in one-engine-inoperative condition
        - OEI_collective_pitch: Collective pitch in one-engine-inoperative condition
        - cruise_tip_mach, cruise_collective_pitch: Additional parameters for prop rotors
    
    Constraints ensure:
        - Thrust and power requirements are met
        - Blade taper is within reasonable bounds (0.3 to 0.9)
        - Blade twist is positive
        - Maximum sectional lift coefficient is below 0.8
        - Chord and twist distribution parameters maintain reasonable ratios
    
    **Major Assumptions**
        * Minimum allowable blade taper: 0.3
        * Maximum allowable blade taper: 0.9
        * Maximum sectional lift coefficient: 0.8
    
    See Also
    --------
    RCAIDE.Library.Methods.Powertrain.Converters.Rotor.Design.blade_geometry_setup
    RCAIDE.Library.Methods.Powertrain.Converters.Rotor.Design.procedure_setup
    """    
    nexus                        = Nexus()
    problem                      = Data()
    nexus.optimization_problem   = problem
   
    if type(rotor) != RCAIDE.Library.Components.Powertrain.Converters.Prop_Rotor or  type(rotor) != RCAIDE.Library.Components.Powertrain.Converters.Lift_Rotor:
        assert('rotor must be of Lift-Rotor or Prop-Rotor class') 
        
    if type(rotor) == RCAIDE.Library.Components.Powertrain.Converters.Prop_Rotor:
        nexus.prop_rotor_flag = True 
    else:
        nexus.prop_rotor_flag = False 
    # -------------------------------------------------------------------
    # Inputs
    # -------------------------------------------------------------------  
    R         = rotor.tip_radius  
    tm_ll_h   = rotor.optimization_parameters.tip_mach_range[0]
    tm_ul_h   = rotor.optimization_parameters.tip_mach_range[1] 
    tm_0_h    = (tm_ul_h + tm_ll_h)/2 
    
    if nexus.prop_rotor_flag:    
        tm_ll_c          = rotor.optimization_parameters.tip_mach_range[0]
        tm_ul_c          = rotor.optimization_parameters.tip_mach_range[1]    
        
    inputs = []
    inputs.append([ 'chord_r'               ,  0.1*R    , 0.05*R     , 0.2*R     , 1.0     ,  1*Units.less])
    inputs.append([ 'chord_p'               ,  2        , 0.25       , 2.0       , 1.0     ,  1*Units.less])
    inputs.append([ 'chord_q'               ,  1        , 0.25       , 1.5       , 1.0     ,  1*Units.less])
    inputs.append([ 'chord_t'               ,  0.05*R   , 0.02*R     , 0.1*R     , 1.0     ,  1*Units.less])  
    inputs.append([ 'twist_r'               ,  np.pi/6  ,  0         , np.pi/4   , 1.0     ,  1*Units.less])
    inputs.append([ 'twist_p'               ,  1        , 0.25       , 2.0       , 1.0     ,  1*Units.less])
    inputs.append([ 'twist_q'               ,  0.5      , 0.25       , 1.5       , 1.0     ,  1*Units.less])
    inputs.append([ 'twist_t'               ,  np.pi/6  , 0          , np.pi/4   , 1.0     ,  1*Units.less])  
    inputs.append([ 'hover_tip_mach'        , tm_0_h    , tm_ll_h    , tm_ul_h   , 1.0     ,  1*Units.less])
    inputs.append([ 'OEI_tip_mach'          , tm_0_h    , tm_ll_h    , 0.85      , 1.0     ,  1*Units.less])
    inputs.append([ 'OEI_collective_pitch'  , np.pi/6   , -np.pi/5   , np.pi/5   , 1.0      ,  1*Units.less])
    if nexus.prop_rotor_flag: 
        inputs.append([ 'cruise_tip_mach'         , tm_ll_c , tm_ll_c    , tm_ul_c  , 1.0     ,  1*Units.less]) 
        inputs.append([ 'cuise_collective_pitch'  , np.pi/8 , -np.pi/5   , np.pi/5  , 1.0     ,  1*Units.less]) 
    problem.inputs = np.array(inputs,dtype=object)   

    # -------------------------------------------------------------------
    # Objective
    # ------------------------------------------------------------------- 
    problem.objective = np.array([  
                                 [  'objective'  ,  1.0   ,    1*Units.less] 
    ],dtype=object)
    
    # -------------------------------------------------------------------
    # Constraints
    # -------------------------------------------------------------------  
    constraints = [] 
    constraints.append([ 'hover_thrust_pow_res'      ,  '<'  ,  1E-3 ,   1.0   , 1*Units.less])  
    constraints.append([ 'blade_taper_constraint_1'  ,  '>'  ,  0.3 ,   1.0   , 1*Units.less])  
    constraints.append([ 'blade_taper_constraint_2'  ,  '<'  ,  0.9 ,   1.0   , 1*Units.less])  
    constraints.append([ 'blade_twist_constraint'    ,  '>'  ,  0.0 ,   1.0   , 1*Units.less])
    constraints.append([ 'max_sectional_cl_hov'      ,  '<'  ,  0.8 ,   1.0   , 1*Units.less])
    constraints.append([ 'chord_p_to_q_ratio'        ,  '>'  ,  0.5 ,   1.0   , 1*Units.less])    
    constraints.append([ 'twist_p_to_q_ratio'        ,  '>'  ,  0.5 ,   1.0   , 1*Units.less])  
    constraints.append([ 'OEI_hov_thrust_pow_res'    ,  '<'  ,  1E-3 ,   1.0   , 1*Units.less]) 
    if nexus.prop_rotor_flag:
        constraints.append([ 'cruise_thrust_pow_res'     ,  '<'  ,  1E-3 ,   1.0   , 1*Units.less]) 
        constraints.append([ 'max_sectional_cl_cruise'   ,  '<'  ,  0.8 ,   1.0   , 1*Units.less])   
    problem.constraints =  np.array(constraints,dtype=object)                
    
    # -------------------------------------------------------------------
    #  Aliases
    # ------------------------------------------------------------------- 
    aliases = [] 
    aliases.append([ 'chord_r'                    , 'vehicle_configurations.*.networks.electric.propulsors.electric_rotor.rotor.chord_r' ])
    aliases.append([ 'chord_p'                    , 'vehicle_configurations.*.networks.electric.propulsors.electric_rotor.rotor.chord_p' ])
    aliases.append([ 'chord_q'                    , 'vehicle_configurations.*.networks.electric.propulsors.electric_rotor.rotor.chord_q' ])
    aliases.append([ 'chord_t'                    , 'vehicle_configurations.*.networks.electric.propulsors.electric_rotor.rotor.chord_t' ]) 
    aliases.append([ 'twist_r'                    , 'vehicle_configurations.*.networks.electric.propulsors.electric_rotor.rotor.twist_r' ])
    aliases.append([ 'twist_p'                    , 'vehicle_configurations.*.networks.electric.propulsors.electric_rotor.rotor.twist_p' ])
    aliases.append([ 'twist_q'                    , 'vehicle_configurations.*.networks.electric.propulsors.electric_rotor.rotor.twist_q' ])
    aliases.append([ 'twist_t'                    , 'vehicle_configurations.*.networks.electric.propulsors.electric_rotor.rotor.twist_t' ])     
    aliases.append([ 'hover_tip_mach'             , 'vehicle_configurations.hover.networks.electric.propulsors.electric_rotor.rotor.hover.design_tip_mach' ]) 
    aliases.append([ 'objective'                  , 'summary.objective'       ])  
    aliases.append([ 'hover_thrust_pow_res'       , 'summary.hover_thrust_power_residual'   ]) 
    aliases.append([ 'blade_taper_constraint_1'   , 'summary.blade_taper_constraint_1'])   
    aliases.append([ 'blade_taper_constraint_2'   , 'summary.blade_taper_constraint_2'])   
    aliases.append([ 'blade_twist_constraint'     , 'summary.blade_twist_constraint'])    
    aliases.append([ 'max_sectional_cl_hov'       , 'summary.max_sectional_cl_hover'])   
    aliases.append([ 'chord_p_to_q_ratio'         , 'summary.chord_p_to_q_ratio'    ])  
    aliases.append([ 'twist_p_to_q_ratio'         , 'summary.twist_p_to_q_ratio'    ])   
    aliases.append([ 'OEI_hov_thrust_pow_res'     , 'summary.oei_thrust_power_residual'   ]) 
    aliases.append([ 'OEI_collective_pitch'       , 'vehicle_configurations.oei.networks.electric.propulsors.electric_rotor.rotor.oei.design_blade_pitch_command' ]) 
    aliases.append([ 'OEI_tip_mach'               , 'vehicle_configurations.oei.networks.electric.propulsors.electric_rotor.rotor.oei.design_tip_mach' ]) 
    if nexus.prop_rotor_flag: 
        aliases.append([ 'cruise_tip_mach'        , 'vehicle_configurations.cruise.networks.electric.propulsors.electric_rotor.rotor.cruise.design_tip_mach' ])  
        aliases.append([ 'cuise_collective_pitch' , 'vehicle_configurations.cruise.networks.electric.propulsors.electric_rotor.rotor.cruise.design_blade_pitch_command' ])  
        aliases.append([ 'cruise_thrust_pow_res'  , 'summary.cruise_thrust_power_residual'   ]) 
        aliases.append([ 'max_sectional_cl_cruise', 'summary.max_sectional_cl_cruise'])  
         
    problem.aliases = aliases
    
    # -------------------------------------------------------------------
    #  Vehicles
    # -------------------------------------------------------------------
    nexus.vehicle_configurations = blade_geometry_setup(rotor,number_of_stations)
    
    # -------------------------------------------------------------------
    #  Analyses
    # -------------------------------------------------------------------
    nexus.analyses = None 
    
    # -------------------------------------------------------------------
    #  Missions
    # -------------------------------------------------------------------
    nexus.missions = None
    
    # -------------------------------------------------------------------
    #  Procedure
    # -------------------------------------------------------------------    
    nexus.print_iterations  = print_iterations 
    nexus.procedure         = procedure_setup()
    
    # -------------------------------------------------------------------
    #  Summary
    # -------------------------------------------------------------------    
    nexus.summary        = Data()     
    nexus.results.hover  = Data() 
    nexus.results.cruise = Data()
    nexus.results.oei    = Data()
    
    return nexus   


# RCAIDE/Library/Attributes/Cryogens/Liquid_Hydrogen.py
# 
#
# Created:  Mar 2024, M. Clarke

# ---------------------------------------------------------------------------------------------------------------------- 
#  Imports
# ---------------------------------------------------------------------------------------------------------------------- 

from .Cryogen import Cryogen

# ---------------------------------------------------------------------------------------------------------------------- 
#  Liquid Hydrogen Cryogen
# ----------------------------------------------------------------------------------------------------------------------  
class Liquid_Hydrogen(Cryogen):
    """
    A class representing liquid hydrogen cryogenic fuel and its thermodynamic properties.

    Attributes
    ----------
    tag : str
        Identifier for the cryogen type ('Liquid_Hydrogen')
    density : float
        Density of liquid hydrogen in kg/m³
    specific_energy : float
        Specific energy content in J/kg
    energy_density : float
        Energy density in J/m³
    temperatures : Container
        Temperature thresholds for the cryogen
            - freeze : float
                Freezing point temperature in Kelvin
            - boiling : float
                Boiling point temperature in Kelvin
    vaporization_enthalpy : float
        Enthalpy of vaporization in kJ/kg
    specific_heat : float
        Specific heat capacity in kJ/kg·K
    liquid_cp_coefficients : Data
        Coefficients for liquid specific heat capacity polynomial
            - LCP_C0 : float
                Constant term coefficient
            - LCP_C1 : float
                Linear term coefficient
            - LCP_C2 : float
                Quadratic term coefficient
            - LCP_C3 : float
                Cubic term coefficient
            - LCP_minT : float
                Minimum valid temperature in Kelvin
            - LCP_maxT : float
                Maximum valid temperature in Kelvin
    
    gas_cp_coefficients : Data
        Coefficients for gas specific heat capacity polynomial
            - GCP_C0 : float
                Constant term coefficient
            - GCP_C1 : float
                Linear term coefficient
            - GCP_C2 : float
                Quadratic term coefficient
            - GCP_C3 : float
                Cubic term coefficient
            - GCP_minT : float
                Minimum valid temperature in Kelvin
            - GCP_maxT : float
                Maximum valid temperature in Kelvin
    
    antoine_coefficients : Data
        Antoine equation coefficients for vapor pressure calculation
            - antoine_A : float
                A coefficient
            - antoine_B : float
                B coefficient
            - antoine_C : float
                C coefficient
            - antoine_minT : float
                Minimum valid temperature in Kelvin
            - antoine_maxT : float
                Maximum valid temperature in Kelvin
    
    vaporization_coefficients : Data
        Coefficients for vaporization enthalpy polynomial
            - H_C0 : float
                Constant term coefficient
            - H_C1 : float
                Linear term coefficient
            - H_C2 : float
                Quadratic term coefficient
            - H_C3 : float
                Cubic term coefficient
            - H_minP : float
                Minimum valid pressure in Pascal
            - H_maxP : float
                Maximum valid pressure in Pascal

    Notes
    -----
    The class implements various thermodynamic properties using polynomial fits and the Antoine equation.
    Specific heat capacity calculations are provided for both liquid and gas phases.
    Valid temperature and pressure ranges are specified for each correlation.
    
    **Definitions**
    
    'Antoine Equation'
        An equation relating vapor pressure to temperature: log10(P) = A - (B/(T+C))
    
    'Specific Heat Capacity'
        The amount of heat required to raise the temperature of 1 kg of the substance by 1 Kelvin
    
    References
    ----------
    [1] Ekin, J. (2006). Experimental techniques for low-temperature measurements: Cryostat design, material properties and superconductor critical-current testing. Oxford University Press. 
    [2] National Institute of Standards and Technology. (2023). NIST Chemistry Webbook, SRD 69. Thermophysical Properties of Fluid Systems. https://webbook.nist.gov/chemistry/fluid/ 
    """

    def __defaults__(self):
        """This sets the default values.
        Assumptions:
            None
        
        Source:
            Ekin - Experimental Techniques for Low Temperature Measurements, ISBN 0-19-857054-6
            NIST Chemistry Webbook
       """ 
        
        self.tag                        = 'Liquid_Hydrogen'
        self.density                    =    59.9            # [kg/m^3] 
        self.specific_energy            =   141.86e6         # [J/kg] 
        self.energy_density             =  8491.0e6          # [J/m^3]
        self.temperatures.freeze        =    13.99           # [K]
        self.temperatures.boiling       =    20.271          # [K]
        self.vaporization_enthalpy      =   461.             # [kJ/kg]
        self.specific_heat              =    10.67           # [kJ/kgK]

        # Coefficiencts for polynomial fit of Liquid Specific Heat Capacity (C_P) curve.
        # C_P = CP_C3*T^3 + CP_C2*T^2 + CP_C1*T^1 + CP_C0*T^0 where C_P is Specific Heat Capacity (J/gK) T is temperature (kelvin).
        # Data from NIST Chemistry Webbook. Pressure is 1.295MPa.
        self.LCP_C0                     =   -31.2
        self.LCP_C1                     =     5.56
        self.LCP_C2                     =    -0.272
        self.LCP_C3                     =     4.76E-03
        
        # Range for which this polynomial fit holds
        self.LCP_minT                   =    15.0              # [K]
        self.LCP_maxT                   =    30.0              # [K]

        # Coefficiencts for polynomial fit of Gas Specific Heat Capacity (C_P) curve.
        # C_P = CP_C3*T^3 + CP_C2*T^2 + CP_C1*T^1 + CP_C0*T^0 where C_P is Specific Heat Capacity (J/gK) T is temperature (kelvin).
        # Data from NIST Chemistry Webbook. Pressure is 0.01 MPa
        self.GCP_C0                     =   10.3
        self.GCP_C1                     =   -7.39E-03
        self.GCP_C2                     =    0.221E-03
        self.GCP_C3                     =   -0.516E-06
        
        # Range for which this polynomial fit holds
        self.GCP_minT                   =   20.0              # [K]
        self.GCP_maxT                   =  300.0              # [K]

        # Antoine Equation Coefficients for calculatating the evaporation temperature.
        # log10(P) = A - (B/(T+C)) where P is vapour pressure (Pa) and T temperature (kelvin).
        # Data from NIST Chemistry Webbook, coefficients converted so as to use pressure in Pa.
        self.antoine_A                  =    8.54314
        self.antoine_B                  =   99.395
        self.antoine_C                  =    7.726
        
        # Range for which Antoine Equation is referenced
        self.antoine_minT               =   21.01             # [K]
        self.antoine_maxT               =   32.27             # [K]

        # Coefficiencts for polynomial fit of vapourisation enthalpy
        # ΔH = H_C3*P^3 + H_C2*P^2 + H_C1*P^1 + H_C0*P^0 where ΔH is vapourisation enthalpy (kJ/kg), P is pressure (Pa). 
        self.H_C0                       =   464.
        self.H_C1                       =    -0.000176
        self.H_C2                       =    52.3E-12
        self.H_C3                       =  -100.00E-18
        
        # Range for which this polynomial fit holds
        self.H_minP                     =     0.02E6         # [Pa]
        self.H_maxP                     =     1.20E6         # [Pa]




# RCAIDE/Library/Methods/Powertrain/Converters/Ram/compute_ram_performance.py
# 
# Created:  Jun 2024, M. Clarke    

# ----------------------------------------------------------------------------------------------------------------------
# compute_ram_performance
# ----------------------------------------------------------------------------------------------------------------------     
def compute_ram_performance(ram, conditions):
    """
    Computes the thermodynamic properties of air at the inlet of a propulsion system.
    
    Parameters
    ----------
    ram : RCAIDE.Library.Components.Converters.Ram
        Ram air converter component with the following attributes:
            - tag : str
                Identifier for the ram air converter
            - working_fluid : Data
                Working fluid properties object
    conditions : RCAIDE.Framework.Mission.Common.Conditions
        Flight conditions with:
            - freestream : Data
                Freestream properties
                    - pressure : numpy.ndarray
                        Freestream static pressure [Pa]
                    - temperature : numpy.ndarray
                        Freestream static temperature [K]
                    - mach_number : numpy.ndarray
                        Freestream Mach number
    
    Returns
    -------
    None
        Results are stored in:
            conditions.freestream:
                - stagnation_temperature : numpy.ndarray
                    Freestream stagnation temperature [K]
                - stagnation_pressure : numpy.ndarray
                    Freestream stagnation pressure [Pa]
                - isentropic_expansion_factor : numpy.ndarray
                    Ratio of specific heats (gamma) [unitless]
                - specific_heat_at_constant_pressure : numpy.ndarray
                    Specific heat at constant pressure [J/(kg·K)]
                - gas_specific_constant : numpy.ndarray
                    Gas specific constant [J/(kg·K)]
                - speed_of_sound : numpy.ndarray
                    Speed of sound [m/s]
            conditions.energy.converters[ram.tag].outputs:
                - stagnation_temperature : numpy.ndarray
                    Stagnation temperature [K]
                - stagnation_pressure : numpy.ndarray
                    Stagnation pressure [Pa]
                - isentropic_expansion_factor : numpy.ndarray
                    Ratio of specific heats (gamma) [unitless]
                - specific_heat_at_constant_pressure : numpy.ndarray
                    Specific heat at constant pressure [J/(kg·K)]
                - gas_specific_constant : numpy.ndarray
                    Gas specific constant [J/(kg·K)]
                - static_temperature : numpy.ndarray
                    Static temperature [K]
                - static_pressure : numpy.ndarray
                    Static pressure [Pa]
                - mach_number : numpy.ndarray
                    Mach number
                - velocity : numpy.ndarray
                    Velocity [m/s]
                - speed_of_sound : numpy.ndarray
                    Speed of sound [m/s]
    
    Notes
    -----
    This function computes the stagnation (total) properties of the air at the inlet
    of a propulsion system based on the freestream conditions. It calculates the
    stagnation temperature and pressure using isentropic flow relations, and also
    computes various thermodynamic properties of the working fluid.
    
    The computation follows these steps:
        1. Extract freestream conditions (pressure, temperature, Mach number)
        2. Compute working fluid properties (gamma, Cp, R, speed of sound)
        3. Calculate freestream velocity from Mach number and speed of sound
        4. Compute stagnation temperature and pressure using isentropic flow relations
        5. Store results in both freestream and ram output conditions
    
    **Major Assumptions**
        * Isentropic flow from freestream to inlet
        * No losses in the inlet
        * Working fluid properties are computed at freestream conditions
    
    **Theory**
    The stagnation temperature and pressure are calculated using the following isentropic relations:
    
    .. math::
        T_0 = T \\left(1 + \\frac{\\gamma - 1}{2} M^2 \\right)
    
    .. math::
        P_0 = P \\left(1 + \\frac{\\gamma - 1}{2} M^2 \\right)^{\\gamma/(\\gamma-1)}
    
    where:
        - :math:`T_0` is the stagnation temperature
        - :math:`T` is the static temperature
        - :math:`P_0` is the stagnation pressure
        - :math:`P` is the static pressure
        - :math:`\\gamma` is the ratio of specific heats
        - :math:`M` is the Mach number
    
    References
    ----------
    [1] Cantwell, B., "AA283 Course Notes", Stanford University https://web.stanford.edu/~cantwell/AA283_Course_Material/AA283_Course_Notes/
    
    See Also
    --------
    RCAIDE.Library.Methods.Powertrain.Converters.Compressor.compute_compressor_performance
    """
    # Unpack flight conditions 
    M0 = conditions.freestream.mach_number
    P0 = conditions.freestream.pressure
    T0 = conditions.freestream.temperature

    # Unpack ram inputs
    working_fluid  = ram.working_fluid
    ram_conditions = conditions.energy.converters[ram.tag]
 
    # Compute the working fluid properties
    R        = working_fluid.gas_specific_constant
    gamma    = working_fluid.compute_gamma(T0,P0) 
    Cp       = working_fluid.compute_cp(T0,P0)
    a        = working_fluid.compute_speed_of_sound(T0,P0)
    V0       = a*M0 

    # Compute the stagnation quantities from the input static quantities
    stagnation_pressure    = P0*((1.+(gamma-1.)/2.*M0*M0 )**(gamma/(gamma-1.))) 
    stagnation_temperature = T0*(1.+((gamma-1.)/2.*M0*M0))

    # Store values into flight conditions data structure  
    conditions.freestream.isentropic_expansion_factor          = gamma
    conditions.freestream.specific_heat_at_constant_pressure   = Cp
    conditions.freestream.gas_specific_constant                = R
    conditions.freestream.stagnation_temperature               = stagnation_temperature
    conditions.freestream.stagnation_pressure                  = stagnation_pressure

    # Store values into compoment outputs  
    ram_conditions.outputs.isentropic_expansion_factor         = gamma
    ram_conditions.outputs.specific_heat_at_constant_pressure  = Cp
    ram_conditions.outputs.gas_specific_constant               = R
    ram_conditions.outputs.stagnation_temperature              = stagnation_temperature
    ram_conditions.outputs.stagnation_pressure                 = stagnation_pressure 
    ram_conditions.outputs.static_temperature                  = T0
    ram_conditions.outputs.static_pressure                     = P0
    ram_conditions.outputs.mach_number                         = M0
    ram_conditions.outputs.velocity                            = V0
    ram_conditions.outputs.speed_of_sound                      = a    
    
    return 


# RCAIDE/Library/Methods/Powertrain/Converters/Turbine/compute_turbine_performance.py
# 
# Created:  Jun 2024, M. Clarke    

# ----------------------------------------------------------------------------------------------------------------------
#  compute_turbine_performance
# ----------------------------------------------------------------------------------------------------------------------     
def compute_turbine_performance(turbine,conditions):
    """
    Computes turbine performance parameters based on input conditions and component characteristics.
    
    Parameters
    ----------
    turbine : RCAIDE.Library.Components.Powertrain.Converters.Turbine
        Turbine component with the following attributes:
            - tag : str
                Identifier for the turbine
            - working_fluid : Data
                Working fluid object with methods to compute properties
                    - compute_gamma : function
                        Computes ratio of specific heats
                    - compute_cp : function
                        Computes specific heat at constant pressure
                    - compute_R : function
                        Computes gas constant
            - mechanical_efficiency : float
                Mechanical efficiency [unitless]
            - polytropic_efficiency : float
                Polytropic efficiency [unitless]
    conditions : RCAIDE.Framework.Mission.Common.Conditions
        Flight conditions with:
            - energy : Data
                Energy conditions
                    - converters : dict
                        Converter energy conditions indexed by tag
                            - inputs : Data
                                Input conditions
                                    - static_temperature : array_like
                                        Static temperature [K]
                                    - static_pressure : array_like
                                        Static pressure [Pa]
                                    - mach_number : array_like
                                        Mach number [unitless]
                                    - stagnation_temperature : array_like
                                        Entering stagnation temperature [K]
                                    - stagnation_pressure : array_like
                                        Entering stagnation pressure [Pa]
                                    - bypass_ratio : array_like
                                        Bypass ratio [unitless]
                                    - fuel_to_air_ratio : array_like
                                        Fuel-to-air ratio [unitless]
                                    - compressor : Data
                                        Compressor data
                                            - work_done : array_like
                                                Compressor work [J/(kg/s)]
                                            - external_shaft_work_done : array_like
                                                Shaft power off take [J/(kg/s)]
                                    - fan : Data
                                        Fan data
                                            - work_done : array_like
                                                Fan work done [J/(kg/s)]
    
    Returns
    -------
    None
        Results are stored in conditions.energy.converters[turbine.tag].outputs:
            - stagnation_pressure : array_like
                Exiting stagnation pressure [Pa]
            - stagnation_temperature : array_like
                Exiting stagnation temperature [K]
            - stagnation_enthalpy : array_like
                Exiting stagnation enthalpy [J/kg]
            - static_temperature : array_like
                Exiting static temperature [K]
            - static_pressure : array_like
                Exiting static pressure [Pa]
            - mach_number : array_like
                Exiting Mach number [unitless]
            - gas_constant : array_like
                Gas constant [J/(kg·K)]
            - pressure_ratio : array_like
                Pressure ratio across turbine [unitless]
            - temperature_ratio : array_like
                Temperature ratio across turbine [unitless]
            - gamma : array_like
                Ratio of specific heats [unitless]
            - cp : array_like
                Specific heat at constant pressure [J/(kg·K)]
    
    Notes
    -----
    This function calculates the performance of a turbine by computing the energy
    extraction required to drive the compressor, fan, and any external power loads.
    It then determines the resulting thermodynamic properties at the turbine exit.
    
    The computation follows these steps:
        1. Extract turbine parameters and working fluid properties
        2. Compute the working fluid properties (gamma, Cp, R) at inlet conditions
        3. Calculate the energy drop across the turbine based on compressor/fan work
           and mechanical efficiency
        4. Compute the exit stagnation temperature, enthalpy, and pressure
        5. Calculate the exit static temperature and pressure
        6. Compute performance ratios (pressure ratio, temperature ratio)
        7. Store all results in the conditions data structure
    
    **Major Assumptions**
        * Constant polytropic efficiency throughout the turbine
        * Constant pressure ratio across the turbine
        * The working fluid behaves as a perfect gas
        * Mechanical losses are accounted for through a constant efficiency factor
    
    **Theory**
    The energy balance across the turbine is:
    
    .. math::
        \\Delta h_t = -\\frac{1}{1+f} \\cdot \\frac{W_{comp} + W_{ext} + \\alpha W_{fan}}{\\eta_{mech}}
    
    where:
        - :math:`\\Delta h_t` is the enthalpy drop across the turbine
        - :math:`f` is the fuel-to-air ratio
        - :math:`W_{comp}` is the compressor work
        - :math:`W_{ext}` is the external shaft work
        - :math:`W_{fan}` is the fan work
        - :math:`\\alpha` is the bypass ratio
        - :math:`\\eta_{mech}` is the mechanical efficiency
    
    The exit stagnation temperature is:
    
    .. math::
        T_{t,out} = T_{t,in} + \\frac{\\Delta h_t}{C_p}
    
    The exit stagnation pressure is calculated using the polytropic efficiency:
    
    .. math::
        P_{t,out} = P_{t,in} \\cdot \\left(\\frac{T_{t,out}}{T_{t,in}}\\right)^{\\frac{\\gamma}{(\\gamma-1)\\eta_{poly}}}
    
    References
    ----------
    [1] Mattingly, J.D., "Elements of Gas Turbine Propulsion", AIAA Education Series, 2005
    
    See Also
    --------
    RCAIDE.Library.Methods.Powertrain.Converters.Compressor.compute_compressor_performance
    RCAIDE.Library.Methods.Powertrain.Converters.Fan.compute_fan_performance
    """              
                             
    # Unpack ram inputs       
    working_fluid      = turbine.working_fluid
    turbine_conditions = conditions.energy.converters[turbine.tag]

    # Compute the working fluid properties
    T0              = turbine_conditions.inputs.static_temperature
    P0              = turbine_conditions.inputs.static_pressure  
    M0              = turbine_conditions.inputs.mach_number   
    gamma           = working_fluid.compute_gamma(T0,P0) 
    Cp              = working_fluid.compute_cp(T0,P0) 
    R               = working_fluid.compute_R(T0,P0)    
    
    #Unpack turbine entering properties 
    eta_mech              = turbine.mechanical_efficiency
    etapolt               = turbine.polytropic_efficiency
    alpha                 = turbine_conditions.inputs.bypass_ratio
    Tt_in                 = turbine_conditions.inputs.stagnation_temperature
    Pt_in                 = turbine_conditions.inputs.stagnation_pressure
    compressor_work       = turbine_conditions.inputs.compressor.work_done
    fan_work              = turbine_conditions.inputs.fan.work_done
    f                     = turbine_conditions.inputs.fuel_to_air_ratio  
    external_power        = turbine_conditions.inputs.compressor.external_shaft_work_done  
  
    # Using the work done by the compressors/fan and the fuel to air ratio to compute the energy drop across the turbine
    deltah_ht = -1/(1+f) * (compressor_work + external_power + alpha * fan_work) * 1/eta_mech
    
    # Compute the output stagnation quantities from the inputs and the energy drop computed above
    Tt_out    = Tt_in+deltah_ht/Cp
    ht_out    = Cp*Tt_out   
    Pt_out    = Pt_in*(Tt_out/Tt_in)**(gamma/((gamma-1)*etapolt)) 
    pi_t      = Pt_out/Pt_in
    tau_t     = Tt_out/Tt_in
    T_out     = Tt_out/(1.+(gamma-1.)/2.*M0*M0)
    P_out     = Pt_out/((1.+(gamma-1.)/2.*M0*M0)**(gamma/(gamma-1.)))         
    
    # Pack outputs of turbine 
    turbine_conditions.outputs.stagnation_pressure     = Pt_out
    turbine_conditions.outputs.stagnation_temperature  = Tt_out
    turbine_conditions.outputs.stagnation_enthalpy     = ht_out
    turbine_conditions.outputs.static_temperature      = T_out
    turbine_conditions.outputs.static_pressure         = P_out 
    turbine_conditions.outputs.mach_number             = M0 
    turbine_conditions.outputs.gas_constant            = R 
    turbine_conditions.outputs.pressure_ratio          = pi_t   
    turbine_conditions.outputs.temperature_ratio       = tau_t     
    turbine_conditions.outputs.gamma                   = gamma 
    turbine_conditions.outputs.cp                      = Cp    
    
    return